/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef CONFIGURATION_H
#define CONFIGURATION_H

#include <libetpan/clist.h>

#include "cntree.h"

#include "folder.h"
#include "mailbox-types.h"
#include "processing.h"
#include "foldersview.h"

#define XEP_CONFIGURATION_FILE_FORMAT_VERSION "0.0.2"

typedef struct XePConfiguration_s XePConfiguration;

struct XePConfiguration_s
{
  clist *identities;
  clist *mtas;
  clist *vfolders;
  XePProcessing *processing;
  XePFoldersViewsList *foldersviewslist;
  chash *folder_refs;
};

/*! Create a configuration structure from an XML tree. 
 *  \param ntree a cntree created when reading an XML configuration file.
 *  \return an allocated configuration structure, or NULL on failure.
 */
XePConfiguration *
a_Configuration_create_from_tree(cntree *ntree);

/*! Create an XML tree from a configuration structure . 
 *  \param configuration the XePConfiguration structure to convert
 *  \return on success, an allocated cntree suitable to be written in an XML 
 *          configuration file; or NULL on failure.
 */
cntree *
a_Configuration_create_tree(XePConfiguration *configuration);

/*! Release resources used by a configuration structure.
 *  \param configuration the configuration strucuture to be deleted.
 */
void
a_Configuration_free(XePConfiguration *configuration);

/*! Link the folder ids in the mailboxes and in the identities to the
 *  corresponding folder
 */
int
a_Configuration_link_folders(XePConfiguration *configuration,
			     XePFoldersList *folderslist);
#endif
