/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef FOLDERSVIEW_H
#define FOLDERSVIEW_H

#include <libetpan/clist.h>

#include "cntree.h"
#include "folder.h"

/*! A folder tree.  
 *
 *  It is a virtual hierarchical organisation of folders.  It is
 *  called virtual because it does not reflect the real folder
 *  hierarchy but rather a tree organisation chosen by the user.  A
 *  folder tree does not necessarily contain all the known
 *  folders.  Each contained folder can be made visible or invisible,
 *  and can have a user chosen path within the tree.
 */

/*! The folder node in the tree view. */
typedef struct XePFoldersViewNode_s XePFoldersViewNode;
struct XePFoldersViewNode_s
{
  XePFolderRef folder_ref; /**< The folder's ID */
  char *name;              /**< The folder's name if it is just an *
			    *   empty node. */
  int property;            /**< Its properties: visible, collapsed. */
  int enabled;
};
 
/*! The folders view. */
typedef struct XePFoldersView_s XePFoldersView;
struct XePFoldersView_s
{
  char *id;                           /**< The name of the view. */
  cntree *tree;                       /**< The tree of folders nodes.  */
  XePFoldersViewNode *lastviewed;     /**< Last viewed node */
};

typedef enum
  {
    XETPAN_FOLDERSVIEW_NODE_VISIBLE   = 1 << 0, /**< Node is visible. */
    XETPAN_FOLDERSVIEW_NODE_COLLAPSED = 1 << 1  /**< Node is collapsed. */
  } XePFoldersViewNodeProperty;

/*! The list of folders views. */
typedef struct XePFoldersViewsList_s XePFoldersViewsList;

struct XePFoldersViewsList_s
{
  clist *list; /**< List of user defined XePFoldersView */
  XePFoldersView *mailboxes; /**< The automatically generated folder view */
};

XePFoldersView *
a_Foldersview_new(void);

void
a_Foldersview_free(XePFoldersView *foldersview);
     
XePFoldersViewNode *
a_Foldersview_node_new(void);

void
a_Foldersview_node_free(XePFoldersViewNode *foldersviewnode);

int
a_Foldersview_mailboxes_view_update(void);

void
a_Foldersview_list_expunge(void);

int
a_Foldersview_list_create(cntree *ntree, 
			  XePFoldersViewsList **foldersviewslist,
			  int op);

XePFoldersViewsList *
a_Foldersview_list_new(void);

void
a_Foldersview_list_free(XePFoldersViewsList *foldersviewslist);

int
a_Foldersview_list_link_folders(XePFoldersViewsList *foldersviewslist);
/*! Return the XePFoldersView which ID is equal to \param foldersview_id
 *  \param foldersviewslist list of virtual trees
 *  \param foldersview_id virtual tree to look for
 *  \return the corresponding XePFoldersView, or the first one if \param
 *  foldersview_id is NULL or empty
 */
XePFoldersView *
a_Foldersview_list_get_foldersview(XePFoldersViewsList *foldersviewslist,
				   const char *foldersview_id);
#endif
