/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This is part of XetPan
 *
 * This is a tree view/list view widget taylored for small memory footprint
 * and for direct accessing of user data via user provided function
 * to retrieve data and to navigate through it.
 * 
 * Some parts of this code are heavily inspired from gtk+ 2.x gtktreeview,
 * for example: resize grips in header button, some function names.
 *
 * Gtk+ is to be found at http://www.gtk.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_GTKUI

#include <gtk/gtksignal.h>
#include <gtk/gtktable.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkviewport.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkarrow.h>
#include <gtk/gtkhscrollbar.h>
#include <gtk/gtkvscrollbar.h>
#include <gtk/gtkrange.h>
#include <gtk/gtktreeview.h>
#include <gtk/gtkstyle.h>
#include <gdk/gdkkeysyms.h>
#include "gtkxtreeview.h"
#include "gtkxdrawingarea.h"

#include <math.h>

#define H_SPACE 12

#undef DEBUG

enum {
	GTK_XTREE_VIEW_NON_NULL     = 1 << 31, /* see cache_set_row_params */
	GTK_XTREE_VIEW_HAS_CHILDREN = 1 << 30,
	GTK_XTREE_VIEW_COLLAPSED    = 1 << 29,
	GTK_XTREE_VIEW_SELECTED     = 1 << 28,
	GTK_XTREE_VIEW_HIDDEN       = 1 << 27,
	GTK_XTREE_VIEW_DEPTH_MASK = ~(
			GTK_XTREE_VIEW_HIDDEN    |
			GTK_XTREE_VIEW_SELECTED  | 
			GTK_XTREE_VIEW_COLLAPSED |
			GTK_XTREE_VIEW_HAS_CHILDREN)
};
#define g_marshal_value_peek_int(v)     (v)->data[0].v_int
#define g_marshal_value_peek_uint(v)    (v)->data[0].v_uint
#define g_marshal_value_peek_pointer(v) (v)->data[0].v_pointer

#define DAREA_DEFAULT_WIDTH 100
#define DAREA_DEFAULT_HEIGHT 100

typedef struct _GtkXTreeViewColumn GtkXTreeViewColumn;

struct _GtkXTreeViewColumn
{
	guint col_id; /* A non zero user defined unique identifier */

	GtkXColumnType type;

	gboolean is_sortable;
	GtkSortType sort_order;

	guint width;

	GtkWidget *widget; /* The button of the title */
	GtkWidget *hbox;   /* the hbox in the title button */
	GtkWidget *user_widget; /* The user provided widget for the title */
	GtkWidget *label;  /* the label of the title button */
	GtkWidget *arrow;  /* the sorting direction arrow in the title button */

	GtkXTreeView *treeview; /* The associated GtkXTreeView */

	GdkWindow *resize_grip; /* to resize */
	gboolean being_resized;
};

struct _GtkXTreeViewClass
{
	GtkTableClass parent_class;

	void (*sort_changed)(GtkXTreeView *xtv, guint col_id, GtkSortType sort_type, gpointer data); /* default handler */
	void (*row_activated)(GtkXTreeView *xtv, gpointer iterator, guint col_id, gpointer data); /* default handler */
	void (*select_cursor_row)(GtkXTreeView *xtv, gpointer iterator, gpointer data); /* default handler */
};

struct _GtkXTreeView
{
	GtkTable table;

	GtkWidget *vbox;
	GtkWidget *hbox; /* hbox containing title buttons */
	GtkWidget *darea; /* drawing area */
	GtkWidget *viewport; /* viewport */
	GtkWidget *hscrollbar; /* horizontal scrolling */
	GtkWidget *vscrollbar; /* scrolls up and down, controls first row iterator */

	guint col_nb;
	guint sort_col_id;
	GtkSortType sort_order;
	guint expander_col_id;
	guint expander_width;
	guint expansible_col_id;

	gboolean default_collapsed;
	gint freeze;

	void *first_iterator;
	void *last_iterator;
	guint last_row;
	guint last_partial;
	uint32_t first_value;

	void *focused_iterator;
	uint32_t focused_value;
	uint32_t total_selected;

	GSList *columns; /* items are GtkXTreeViewColumn* */

	guint allocated_rows;
	guint allocated_cols;

	void ***cells; /* cells[0..rows][0..cols] are pointer to PangoLayout or 
				Pixbufs */
	guint *cell_prop;
	guint *cell_prop_tmp;
	void **iterators;
	void **iterators_tmp;
	gint *y_pos;
	gint *y_pos_tmp;
	guint rows;
	guint displayed;
	guint length;

	gint y0; /* offset of the first row (negative value) */
	gboolean adjust_first; /* if we need to scroll up when resizing */
	/* User callback functions */
	GtkXTreeViewRowDataFunc row_data_func;
	gpointer row_data_data;

	GtkXTreeViewRowParamsFunc row_params_func;
	gpointer row_params_data;

	GtkXTreeViewDataIteratorFunc data_iterator_func;
	gpointer data_iterator_data;

	GtkXTreeViewMatchFunc match_func;
	gpointer match_data;

	GHashTable *cache;
};

enum {
	SORT_CHANGED_SIGNAL,
	ROW_ACTIVATED_SIGNAL,
	SELECT_CURSOR_ROW_SIGNAL,
	LAST_SIGNAL
};

static guint gtkxtreeview_signals[LAST_SIGNAL] = {0};


/**/

static void
_g_cclosure_marshal_VOID__UINT_INT (GClosure     *closure,
                               GValue       *return_value,
                               guint         n_param_values,
                               const GValue *param_values,
                               gpointer      invocation_hint,
                               gpointer      marshal_data);
static void
_g_cclosure_marshal_VOID__POINTER_UINT (GClosure     *closure,
                               GValue       *return_value,
                               guint         n_param_values,
                               const GValue *param_values,
                               gpointer      invocation_hint,
                               gpointer      marshal_data);
static void
_g_cclosure_marshal_VOID__POINTER (GClosure     *closure,
                               GValue       *return_value,
                               guint         n_param_values,
                               const GValue *param_values,
                               gpointer      invocation_hint,
                               gpointer      marshal_data);

static void gtk_xtree_view_class_init(GtkXTreeViewClass *klass);
static void gtk_xtree_view_init(GtkXTreeView *xtv);
static void gtk_xtree_view_style_set(GtkWidget *widget, GtkStyle *previous_style);
static void darea_size_request_cb(GtkWidget *widget, 
		GtkRequisition *requisition);
static void darea_size_allocate_cb(GtkWidget *widget, 
		GtkAllocation *allocation, 
		gpointer axtv);
static gboolean darea_configure_event_cb(GtkWidget *widget, 
		GdkEventConfigure *event, 
		gpointer axtv);
static gboolean darea_expose_event_cb(GtkWidget *widget, 
		GdkEventExpose *event, 
		gpointer axtv);
static void scroll_cb(GtkRange *range, 
		gpointer axtv);

static gboolean darea_fill(GtkXTreeView *xtv, 
		void *firs_iterator, 
		guint first_row, 
		guint last_row);
static void darea_cursor_move_home(GtkXTreeView *xtv, GdkModifierType modifier);
static void darea_cursor_move_end(GtkXTreeView *xtv, GdkModifierType modifier);
static void darea_cursor_scroll(GtkXTreeView *xtv, gboolean is_next, gboolean page,
		GdkModifierType modifier);
static void darea_cursor_move_to(GtkXTreeView *xtv, uint32_t new_value,
		GdkModifierType modifier);

static void cache_set_row_params(GtkXTreeView *xtv, void *iterator, 
		guint params);
static void cache_get_row_params(GtkXTreeView *xtv, void *iterator, 
		guint *params);
static void add_row(GtkXTreeView *xtv);
static void create_title_button(GtkXTreeViewColumn *column, 
		GtkWidget *title_widget, 
		gchar *title_string);

static void gtk_xtree_view_update_adjustment(GtkXTreeView *xtv);
static void update_resize_grips(GtkXTreeView *xtv);
static void header_realize_cb(GtkWidget *button, 
		gpointer acolumn);
static void header_clicked_cb(GtkButton *button, 
		gpointer acolumn);
static gboolean header_scroll_event_cb(GtkWidget *widget, 
		GdkEvent *event,
		gpointer acolumn);
static gboolean header_button_press_cb(GtkWidget *widget, 
		GdkEventButton *event, gpointer acolumn);
static gboolean header_button_release_cb(GtkWidget *widget, 
		GdkEventButton *event, 
		gpointer column);
static gboolean header_event_cb(GtkWidget *widget, 
		GdkEvent *event, 
		gpointer acolumn);

static void xtv_grab_focus_cb(GtkWidget *widget, gpointer axtv);
static gboolean darea_focus_in_cb(GtkWidget *widget, 
		GdkEventFocus *event, 
		gpointer axtv);

static gboolean darea_focus_out_cb(GtkWidget *widget, 
		GdkEventFocus *event, 
		gpointer axtv);

static gboolean darea_button_release_cb(GtkWidget *widget, 
		GdkEventButton *event,
		gpointer axtv);
static gboolean darea_scroll_cb(GtkWidget *widget, 
		GdkEvent *event, 
		gpointer axtv);
static gboolean darea_button_press_cb(GtkWidget *widget, 
		GdkEventButton *event,
		gpointer axtv);
static void gtk_xtree_view_scroll(GtkXTreeView *xtv,
		gboolean is_next, gboolean full_page);
static gboolean gtk_xtree_view_key_press_cb(GtkWidget *widget,
		GdkEventKey *event,
		gpointer axtv);
static gboolean default_match_func(void *iterator, gpointer data);
static gboolean select_iterator(GtkXTreeView *xtv, void *iterator,
		gboolean toggle);
static gboolean expand_node(GtkXTreeView *xtv, void *iterator, guint params);
static gboolean collapse_node(GtkXTreeView *xtv, void *iterator, guint params);

static void
_g_cclosure_marshal_VOID__UINT_INT (GClosure     *closure,
                               GValue       *return_value,
                               guint         n_param_values,
                               const GValue *param_values,
                               gpointer      invocation_hint,
                               gpointer      marshal_data)
{
  typedef void (*GMarshalFunc_VOID__UINT_INT) (gpointer     data1,
		  			   guint        arg_1,
                                           gint         arg_2,
                                           gpointer     data2);
  register GMarshalFunc_VOID__UINT_INT callback;
  register GCClosure *cc = (GCClosure*) closure;
  register gpointer data1, data2;

  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }
  callback = (GMarshalFunc_VOID__UINT_INT) (marshal_data ? marshal_data : cc->callback);

  callback (data1,
            g_marshal_value_peek_uint (param_values + 1),
            g_marshal_value_peek_int (param_values + 2),
            data2);
}

static void
_g_cclosure_marshal_VOID__POINTER_UINT (GClosure     *closure,
                               GValue       *return_value,
                               guint         n_param_values,
                               const GValue *param_values,
                               gpointer      invocation_hint,
                               gpointer      marshal_data)
{
  typedef void (*GMarshalFunc_VOID__POINTER_UINT) (gpointer     data1,
		  			   gpointer        arg_1,
                                           guint         arg_2,
                                           gpointer     data2);
  register GMarshalFunc_VOID__POINTER_UINT callback;
  register GCClosure *cc = (GCClosure*) closure;
  register gpointer data1, data2;

  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }
  callback = (GMarshalFunc_VOID__POINTER_UINT) (marshal_data ? marshal_data : cc->callback);

  callback (data1,
            g_marshal_value_peek_pointer (param_values + 1),
            g_marshal_value_peek_uint (param_values + 2),
            data2);
}

static void
_g_cclosure_marshal_VOID__POINTER (GClosure     *closure,
                               GValue       *return_value,
                               guint         n_param_values,
                               const GValue *param_values,
                               gpointer      invocation_hint,
                               gpointer      marshal_data)
{
  typedef void (*GMarshalFunc_VOID__POINTER) (gpointer     data1,
		  			   gpointer        arg_1,
                                           gpointer     data2);
  register GMarshalFunc_VOID__POINTER callback;
  register GCClosure *cc = (GCClosure*) closure;
  register gpointer data1, data2;

  g_return_if_fail (n_param_values == 2);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }
  callback = (GMarshalFunc_VOID__POINTER) (marshal_data ? marshal_data : cc->callback);

  callback (data1,
            g_marshal_value_peek_pointer (param_values + 1),
            data2);
}

GType gtk_xtree_view_get_type(void)
{
	static GType xtv_type = 0;

	if (!xtv_type) {
		static const GTypeInfo xtv_info = 
		{
			sizeof(GtkXTreeViewClass),
			NULL, /* base_init */
			NULL, /* base_finalize */
			(GClassInitFunc) gtk_xtree_view_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof(GtkXTreeView ),
			0, /* n_preallocs */
			(GInstanceInitFunc) gtk_xtree_view_init
		};

		xtv_type = g_type_register_static(GTK_TYPE_TABLE, "GtkXTreeView", &xtv_info, 0);
	}

	return xtv_type;
}

static void gtk_xtree_view_class_init(GtkXTreeViewClass *klass)
{
	GtkWidgetClass *widget_class = (GtkWidgetClass *) klass;

	widget_class->style_set = gtk_xtree_view_style_set;

	gtkxtreeview_signals[SORT_CHANGED_SIGNAL] = g_signal_new("sort-changed",
			G_TYPE_FROM_CLASS(klass),
			G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
			G_STRUCT_OFFSET(GtkXTreeViewClass, sort_changed),
			NULL, NULL,
			_g_cclosure_marshal_VOID__UINT_INT,
			G_TYPE_NONE,
			2,
			G_TYPE_UINT,
			G_TYPE_INT);
	gtkxtreeview_signals[ROW_ACTIVATED_SIGNAL] = g_signal_new("row-activated",
			G_TYPE_FROM_CLASS(klass),
			G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
			G_STRUCT_OFFSET(GtkXTreeViewClass, row_activated),
			NULL, NULL,
			_g_cclosure_marshal_VOID__POINTER_UINT,
			G_TYPE_NONE,
			2,
			G_TYPE_POINTER,
			G_TYPE_UINT);
	gtkxtreeview_signals[SELECT_CURSOR_ROW_SIGNAL] = g_signal_new("select-cursor-row",
			G_TYPE_FROM_CLASS(klass),
			G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
			G_STRUCT_OFFSET(GtkXTreeViewClass, select_cursor_row),
			NULL, NULL,
			_g_cclosure_marshal_VOID__POINTER,
			G_TYPE_NONE,
			1,
			G_TYPE_POINTER,
			G_TYPE_INT);
}

static void gtk_xtree_view_init(GtkXTreeView *xtv)
{
	gint attach;
	GtkStyle *style;
	gint i;

	GTK_WIDGET_SET_FLAGS(GTK_WIDGET(xtv), GTK_CAN_FOCUS);
	xtv->col_nb = 0;
	xtv->sort_col_id = 0;
	xtv->sort_order = GTK_SORT_ASCENDING;
	xtv->expander_col_id = 0;
	xtv->expander_width = 10;
	xtv->expansible_col_id = 0;
	xtv->default_collapsed = FALSE;
	xtv->freeze = 0;
	xtv->first_iterator = NULL;
	xtv->last_iterator = NULL;
	xtv->last_row = 0;
	xtv->last_partial = 0;
	xtv->first_value = 0;

	xtv->focused_iterator = NULL;
	xtv->focused_value = 0;
	xtv->total_selected = 0;

	xtv->columns = NULL;

	xtv->allocated_rows = 0;
	xtv->allocated_cols = 0;

	xtv->cells = NULL;
	xtv->cell_prop = NULL;
	xtv->cell_prop_tmp = NULL;
	xtv->y_pos = NULL;
	xtv->y_pos_tmp = NULL;
	xtv->iterators = NULL;
	xtv->iterators_tmp = NULL;

	xtv->row_data_func = NULL;
	xtv->row_data_data = NULL;
	xtv->row_params_func = NULL;
	xtv->row_params_data = NULL;
	xtv->data_iterator_func = NULL;
	xtv->data_iterator_data = NULL;
	gtk_xtree_view_set_match_func(xtv, NULL, NULL);

	xtv->y0 = 0;

	xtv->cache = g_hash_table_new(NULL, NULL);

	attach = (gtk_widget_get_direction(GTK_WIDGET(xtv)) == GTK_TEXT_DIR_RTL? 1 : 0);

	gtk_table_resize(GTK_TABLE(xtv), 2, 2);

	xtv->hscrollbar = gtk_hscrollbar_new(NULL);
	gtk_table_attach(GTK_TABLE(xtv), xtv->hscrollbar, attach, attach + 1, 1, 2, GTK_FILL | GTK_EXPAND, GTK_SHRINK, 0 ,0);
	/*gtk_widget_show(xtv->hscrollbar);*/

	xtv->vscrollbar = gtk_vscrollbar_new(NULL);
	gtk_table_attach(GTK_TABLE(xtv), xtv->vscrollbar, 1 - attach, 2 - attach, 0, 1, 
			GTK_SHRINK, GTK_FILL | GTK_EXPAND, 0 ,0);
	gtk_widget_show(xtv->vscrollbar);

	xtv->viewport = gtk_viewport_new(gtk_range_get_adjustment(GTK_RANGE(xtv->hscrollbar)), NULL);
	gtk_table_attach(GTK_TABLE(xtv), xtv->viewport, attach, attach + 1, 0, 1, 
			GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);

	xtv->vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(xtv->viewport), xtv->vbox);

	xtv->hbox = gtk_hbox_new(FALSE, 0);
	gtk_widget_show(xtv->hbox);
	gtk_box_pack_start(GTK_BOX(xtv->vbox), xtv->hbox, FALSE, FALSE, 0);

	xtv->darea = gtk_xdrawing_area_new();
	GTK_WIDGET_SET_FLAGS(xtv->darea, GTK_CAN_FOCUS);
	gtk_widget_add_events(xtv->darea, GDK_BUTTON_PRESS_MASK |
			GDK_BUTTON_RELEASE_MASK |
			GDK_KEY_PRESS_MASK |
			GDK_KEY_RELEASE_MASK |
			GDK_PROPERTY_CHANGE_MASK);
	gtk_box_pack_start(GTK_BOX(xtv->vbox), xtv->darea, TRUE, TRUE, 0);
	gtk_widget_show(xtv->darea);

	gtk_widget_show(xtv->viewport);
	gtk_widget_show(xtv->vbox);

	gtk_xtree_view_set_length(xtv, 0);
#if 0
	g_signal_connect(G_OBJECT(xtv->darea), "size-request",
			 G_CALLBACK(darea_size_request_cb), xtv);
#endif
	g_signal_connect(G_OBJECT(xtv->vbox), "size-allocate",
			 G_CALLBACK(darea_size_allocate_cb), xtv);
	g_signal_connect(G_OBJECT(xtv->darea), "configure-event",
			G_CALLBACK(darea_configure_event_cb), xtv);
	g_signal_connect(G_OBJECT(xtv->darea), "expose-event",
			G_CALLBACK(darea_expose_event_cb), xtv);
	g_signal_connect(G_OBJECT(xtv->vscrollbar), "value-changed",
			G_CALLBACK(scroll_cb), xtv);
	g_signal_connect(G_OBJECT(xtv->darea), "button-press-event",
			G_CALLBACK(darea_button_press_cb), xtv);
	g_signal_connect(G_OBJECT(xtv->darea), "button-release-event",
			G_CALLBACK(darea_button_release_cb), xtv);
	g_signal_connect(G_OBJECT(xtv->darea), "scroll-event",
			G_CALLBACK(darea_scroll_cb), xtv);
	g_signal_connect(G_OBJECT(xtv), "key-press-event",
			G_CALLBACK(gtk_xtree_view_key_press_cb), xtv);
	g_signal_connect(G_OBJECT(xtv), "grab-focus",
			G_CALLBACK(xtv_grab_focus_cb), xtv);
	g_signal_connect(G_OBJECT(xtv->darea), "focus-in-event",
			G_CALLBACK(darea_focus_in_cb), xtv);
	g_signal_connect(G_OBJECT(xtv->darea), "focus-out-event",
			G_CALLBACK(darea_focus_out_cb), xtv);
}

static void gtk_xtree_view_style_set(GtkWidget *widget, GtkStyle *previous_style)
{
	GtkXTreeView *xtv = GTK_XTREE_VIEW(widget);
	GtkWidget *darea = xtv->darea;
	gint row, col;
	GSList *iter;
	GtkXTreeViewColumn *column;
	void **row_cells;
	gchar *font;
	GtkStyle *style;
	GtkWidget *button;

	for (iter = xtv->columns; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		button = column->widget;
		style = gtk_rc_get_style_by_paths(gtk_widget_get_settings(button), NULL,
				"*.GtkTreeView.GtkButton", G_OBJECT_TYPE(button));
		gtk_widget_set_style(button, style);
	}
		
	for (row = 0; row < xtv->allocated_rows; row++) {
		row_cells = xtv->cells[row];
		for (col = 0, iter = xtv->columns; iter; iter = iter->next, col++) {
			column = (GtkXTreeViewColumn *) iter->data;
			switch (column->type) {
				case GTK_XCOLUMN_TYPE_TEXT:
					g_object_unref (row_cells[col]);
					row_cells[col] = gtk_widget_create_pango_layout(darea, NULL);
					break;
				default:
					break;
			}
		}
	}
}

static void xtv_grab_focus_cb(GtkWidget *widget, gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
	gtk_widget_grab_focus(xtv->darea);
}

static gboolean darea_focus_in_cb(GtkWidget *widget, GdkEventFocus *event, gpointer axtv)
{
	GTK_WIDGET_SET_FLAGS(widget, GTK_HAS_FOCUS);
	return FALSE;
}

static gboolean darea_focus_out_cb(GtkWidget *widget, GdkEventFocus *event, gpointer axtv)
{
	GTK_WIDGET_UNSET_FLAGS(widget, GTK_HAS_FOCUS);
	return FALSE;
}
static void
darea_size_request_cb(GtkWidget *widget, GtkRequisition *requisition)
{
#ifdef DEBUG
  g_message("Size request...\n");
#endif
  requisition->width = DAREA_DEFAULT_WIDTH;
  requisition->height = DAREA_DEFAULT_HEIGHT;
}

static void
darea_size_allocate_cb(GtkWidget *widget, 
		GtkAllocation *allocation, 
		gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
	update_resize_grips(xtv);
}

static gboolean
darea_configure_event_cb(GtkWidget *widget, GdkEventConfigure *event, gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
#ifdef DEBUG
	g_message("Configure x %d, y %d, w %d, h %d\n", event->x, event->y, event->width, event->height);
#endif

	if (xtv->freeze == 0) {
		guint old_first_value = xtv->first_value;
		do {
			darea_fill(xtv, xtv->first_iterator, 0, 0);
			if (xtv->adjust_first)
				xtv->adjust_first = 
					xtv->data_iterator_func(&xtv->first_iterator, 
							FALSE, 
							FALSE, 
							xtv->sort_col_id, 
							xtv->sort_order, 
							xtv->data_iterator_data);
			if (xtv->adjust_first)
				xtv->first_value--;
		} while (xtv->adjust_first);
		if (old_first_value != xtv->first_value) {
			GtkRange *range = GTK_RANGE(xtv->vscrollbar);
			gtk_range_set_value(range, xtv->first_value);
		}
	}
	/*gtk_xtree_view_update_adjustment(xtv);*/
		
	return TRUE;
}

static gboolean darea_scroll_cb(GtkWidget *widget, 
		GdkEvent *event, 
		gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
	GdkEventScroll *scroll = (GdkEventScroll *) event;
	
	gtk_xtree_view_scroll(xtv, scroll->direction == GDK_SCROLL_DOWN, FALSE);

	return FALSE;
}

void gtk_xtree_view_scroll_to(GtkXTreeView *xtv, double value)
{
  GtkAdjustment *adj = 
    gtk_range_get_adjustment(GTK_RANGE(xtv->vscrollbar));
  gdouble m;
  adj->upper = (xtv->displayed > 0 ? xtv->displayed : 1.0);

  m = adj->upper - adj->page_increment; 
  printf ("m = %f, v = %f\n", m, value);
  value = ( value >= m ) ? m : value;

  gtk_range_set_value(GTK_RANGE(xtv->vscrollbar), value);
  gtk_xtree_view_update_adjustment(xtv);
}

static void gtk_xtree_view_scroll(GtkXTreeView *xtv,
		gboolean is_next, gboolean full_page)
{
	GtkAdjustment *adj = 
		gtk_range_get_adjustment(GTK_RANGE(xtv->vscrollbar));
	gdouble delta;
	delta = adj->page_increment; 
	if (full_page)
		delta = delta - 1;
	else
		delta = delta / 2;
	if (!is_next)
		delta = -delta;
	gtk_range_set_value(GTK_RANGE(xtv->vscrollbar), adj->value + delta);
}

static gboolean gtk_xtree_view_key_press_cb(GtkWidget *widget,
		GdkEventKey *event,
		gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
	const guint modifier = event->state & GDK_MODIFIER_MASK;
	gboolean rtl = (gtk_widget_get_direction(GTK_WIDGET(xtv)) == 
			GTK_TEXT_DIR_RTL);

	if (xtv->freeze)
		return FALSE;
	if (!GTK_WIDGET_HAS_FOCUS(xtv->darea))
		gtk_widget_grab_focus(xtv->darea);

	switch (event->keyval) {
		case GDK_a:
			if (modifier == GDK_CONTROL_MASK)
				gtk_xtree_view_select_all(xtv, NULL);
			else if (!modifier)
				gtk_xtree_view_select_all_matches(xtv, FALSE);
			break;
		case GDK_i:
			if (modifier == GDK_CONTROL_MASK)
				gtk_xtree_view_invert_selection(xtv, NULL);
			break;
		case GDK_Left:
			if (!modifier) {
				if (rtl)
					gtk_xtree_view_next_match(xtv, FALSE);
				else
					gtk_xtree_view_prev_match(xtv, FALSE);
			}
			break;
		case GDK_Right:
			if (!modifier) {
				if (rtl)
					gtk_xtree_view_prev_match(xtv, FALSE);
				else
					gtk_xtree_view_next_match(xtv, FALSE);
			}
			break;
		case GDK_g:
			if (modifier == GDK_CONTROL_MASK) {
				gtk_xtree_view_next_match(xtv, TRUE);
			} else if (modifier == (GDK_CONTROL_MASK | GDK_SHIFT_MASK)) {
				gtk_xtree_view_prev_match(xtv, TRUE);
			}
			break;
		case GDK_Page_Down:	
		case GDK_Page_Up:
			if (event->state & GDK_MOD1_MASK)
				gtk_xtree_view_scroll(xtv, 
						(event->keyval == GDK_Page_Down),
						TRUE);
			else if ((event->state & GDK_MODIFIER_MASK) == 0)
				darea_cursor_scroll(xtv, 
						(event->keyval == GDK_Page_Down), 
						TRUE,
						event->state);
			break;
		case GDK_Down:
		case GDK_Up:
			if (event->state & GDK_MOD1_MASK)
				gtk_xtree_view_scroll(xtv, 
						(event->keyval == GDK_Down),
						FALSE);
			else if ((event->state & GDK_MODIFIER_MASK) == 0)
				darea_cursor_scroll(xtv,
						(event->keyval == GDK_Down),
						FALSE,
						event->state);
			break;
		case GDK_Home:
			if (event->state & GDK_MOD1_MASK) {
				gtk_range_set_value(GTK_RANGE(xtv->vscrollbar),
						0.0);
			} else if ((event->state & GDK_MODIFIER_MASK) == 0) {
				darea_cursor_move_home(xtv, event->state);
			}
			break;
		case GDK_End:	
			if (event->state & GDK_MOD1_MASK) {
				GtkAdjustment *adj;
				adj = gtk_range_get_adjustment(GTK_RANGE(xtv->vscrollbar));
				gtk_range_set_value(GTK_RANGE(xtv->vscrollbar),
						adj->upper);
			} else if ((event->state & GDK_MODIFIER_MASK) == 0) {
				darea_cursor_move_end(xtv, event->state);
			}
			break;
		case GDK_space:
		case GDK_Return:
			gtk_xtree_view_activate_focused(xtv, event->state);
	}
	return FALSE;
}

static void darea_cursor_move_home(GtkXTreeView *xtv, GdkModifierType modifier)
{
	darea_cursor_move_to(xtv, 0, modifier);
}

static void darea_cursor_move_end(GtkXTreeView *xtv, GdkModifierType modifier)
{
	darea_cursor_move_to(xtv, xtv->displayed - 1, modifier);
}

static void darea_cursor_scroll(GtkXTreeView *xtv, gboolean is_next, gboolean page,
		GdkModifierType modifier)
{
	uint32_t increment;
	uint32_t new_value;

	if (!page) {
		increment = 1;
	} else {
		GtkAdjustment *adj;
		adj = gtk_range_get_adjustment(GTK_RANGE(xtv->vscrollbar));
		increment = adj->page_increment - 1;
	}
	if (is_next) {
		if (xtv->focused_value + increment < xtv->displayed)
			new_value = xtv->focused_value + increment;
		else {
			new_value = xtv->displayed - 1;
			increment = new_value - xtv->focused_value;
		}
	} else {
		if (xtv->focused_value > increment)
			new_value = xtv->focused_value - increment;
		else {
			new_value = 0;
			increment = xtv->focused_value;
		}
	}
	darea_cursor_move_to(xtv, new_value, modifier);
}

static void darea_cursor_move_to(GtkXTreeView *xtv, uint32_t new_value,
		GdkModifierType modifier)
{
	uint32_t increment;
	gboolean is_next;

	if (xtv->focused_value <= new_value) {
		increment = new_value - xtv->focused_value;
		is_next = TRUE;
	} else {
		increment = xtv->focused_value - new_value;
		is_next = FALSE;
	}

	/* Set focused iterator */
	if (xtv->first_value <= new_value && 
			new_value < xtv->first_value + xtv->rows - xtv->last_partial) {
		xtv->focused_iterator = xtv->iterators[new_value - xtv->first_value];
		xtv->focused_value = new_value;
		gtk_xtree_view_update(xtv);
	} else {
		void *iterator;
		/* Outside view => scroll */
		/* This is a quasi duplicate work: scroll does something similar
		 */
		/* We start searching for the iterator from the viewed
		 * boundary
		 */
		if (new_value < xtv->first_value) {
			increment = xtv->first_value - new_value;
			iterator = xtv->first_iterator;
		} else {
			increment = new_value - 
				(xtv->first_value + xtv->rows - xtv->last_partial
				 - 1);
			iterator =
				xtv->iterators[xtv->rows - xtv->last_partial - 1];
		}
		if (is_next) {
			guint params;
			uint32_t i;

			for (i = 0; i < increment; i++) {
			cache_get_row_params(xtv, iterator, &params);
			if (!xtv->data_iterator_func(&iterator,
					TRUE,
					params & GTK_XTREE_VIEW_COLLAPSED,
					xtv->sort_col_id,
					xtv->sort_order,
					xtv->data_iterator_data))
				break;
			}
		} else {
			guint params = 0; /* first iterator is never hidden */
			uint32_t i;
			for (i = 0; i < increment;) {
				if (!xtv->data_iterator_func(&iterator,
							FALSE,
							params & GTK_XTREE_VIEW_HIDDEN,
							xtv->sort_col_id,
							xtv->sort_order,
							xtv->data_iterator_data))
					break;
				cache_get_row_params(xtv, iterator, &params);
				if ((params & GTK_XTREE_VIEW_HIDDEN) == 0)
						i++;
			}
		}
		xtv->focused_iterator = iterator;
		xtv->focused_value = new_value;
		/* new first iterator value to scroll there: */
		if (is_next) {
			new_value -= xtv->rows - xtv->last_partial - 1;
		} 
		gtk_range_set_value(GTK_RANGE(xtv->vscrollbar), new_value);
	}
}

void gtk_xtree_view_activate_focused(GtkXTreeView *xtv, guint state)
{
	gboolean ctrl = (state & GDK_CONTROL_MASK) != 0;
	uint32_t value = xtv->focused_value;
	void *iterator = xtv->focused_iterator;
	gboolean need_scroll;

	if (!ctrl)
		gtk_xtree_view_clear_selection(xtv);
	select_iterator(xtv, iterator, FALSE);
	if (xtv->first_value <= value &&
			value - xtv->first_value <= xtv->last_row) {
		guint index = value - xtv->first_value;
		xtv->cell_prop[index] |= GTK_XTREE_VIEW_SELECTED;
		need_scroll = FALSE;
	} else {
		need_scroll = TRUE;
	}

	if (need_scroll) {
		GtkAdjustment *adj;
		adj = gtk_range_get_adjustment(GTK_RANGE(xtv->vscrollbar));
		gtk_range_set_value(GTK_RANGE(xtv->vscrollbar),
				value - adj->page_increment * 0.5 + 1);
	}
	
	g_signal_emit(G_OBJECT(xtv),
			gtkxtreeview_signals[ROW_ACTIVATED_SIGNAL],	
			0,
			iterator,
			0);
}

static gboolean darea_button_press_cb(GtkWidget *widget, 
		GdkEventButton *event,
		gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
	void *selected = NULL;
	void *to = NULL;
	uint32_t index;
	gboolean ctrl;
	gboolean expander;
	gint row = 0;
	gint col_id = 0;
	gboolean found_cell;
	GdkRectangle rect;
	guint old_focused_value = xtv->focused_value;
	void *old_focused_iterator = xtv->focused_iterator;
	gboolean focused_selected = TRUE;

	if (xtv->freeze)
		return FALSE;
	gtk_widget_grab_focus(xtv->darea);
	
	if (event->button == 1) {
		/* button1: delete old selection and select new row
		 * shift + button1: delete old selection and 
		 *                  select from last focus to new focus 
		 * ctrl  + button1: add to selection
		 */
	
		found_cell = gtk_xtree_view_find_cell(xtv,
				(int) event->x, (int) event->y,
				&selected,
				&row, &col_id, &expander);

		/* if no cell found, ignore click */
		if (!found_cell)
			return FALSE;

		/* Check if user clicked on a expander, in which case we do not
		 * focus or delete the toggled node
		 */
		if (expander) {
			gtk_xtree_view_toggle_node(xtv, selected);
			return FALSE;
		}

		ctrl = (event->state & GDK_CONTROL_MASK) != 0;
		if (!ctrl)
			gtk_xtree_view_clear_selection(xtv);

		xtv->focused_iterator = selected;
		xtv->focused_value = xtv->first_value + row;

		to = selected;
		index = xtv->focused_value;

		if (event->state & GDK_SHIFT_MASK) {
			if (xtv->focused_value > old_focused_value) {
				to = selected;
				selected = old_focused_iterator;
				index = old_focused_value;
			} else {
				to = old_focused_iterator;
			}
		}

		/* Select newly focused item */
		do {
			gboolean toggle = (selected == xtv->focused_iterator) && ctrl;
			if (select_iterator(xtv, selected, toggle)) {
				if ((xtv->first_value <= index) &&
						(index - xtv->first_value <= xtv->last_row)) {
					if (toggle) {
						xtv->cell_prop[index - xtv->first_value] ^= GTK_XTREE_VIEW_SELECTED;
						focused_selected = xtv->cell_prop[index - xtv->first_value] & GTK_XTREE_VIEW_SELECTED;
					}
					else
						xtv->cell_prop[index - xtv->first_value] |= GTK_XTREE_VIEW_SELECTED;
				}
			}
			index++;
		} while (selected != to && 
				xtv->data_iterator_func(&selected,
					TRUE, 
					FALSE, 
					xtv->sort_col_id, 
					xtv->sort_col_id, 
					xtv->data_iterator_data));

		rect.x = 0;
		rect.y = 0;
		rect.width = xtv->darea->allocation.width;
		rect.height = xtv->darea->allocation.height;
		gdk_window_invalidate_rect(xtv->darea->window, &rect, FALSE);
		gdk_window_process_updates(xtv->darea->window, FALSE);

		/* What should be done : 
		 * button 1: select, focus + activate
		 * if Ctrl + Button 1 deselect focused item, do not send activate signal
		 */
		if (focused_selected) {
			g_signal_emit(G_OBJECT(xtv),
					gtkxtreeview_signals[ROW_ACTIVATED_SIGNAL],	
					0,
					xtv->focused_iterator,
					col_id);
		}
	}
	return FALSE;
}

static gboolean select_iterator(GtkXTreeView *xtv, void *iterator, 
		gboolean toggle)
{
	guint params;

	cache_get_row_params(xtv, iterator, &params);
	if (!toggle && (params & GTK_XTREE_VIEW_SELECTED))
		return FALSE;
	if (toggle && (params & GTK_XTREE_VIEW_SELECTED)) {
		/* deselect */
		params ^= GTK_XTREE_VIEW_SELECTED;
		xtv->total_selected--;
	} else {
		params |= GTK_XTREE_VIEW_SELECTED;
		xtv->total_selected++;
	}
	cache_set_row_params(xtv, iterator, params);
	return TRUE;
}

gboolean gtk_xtree_view_find_cell(GtkXTreeView *xtv, 
		gint x, gint y, 
		void **iterator, 
		gint *row, gint *col_id,
		gboolean *expander)
{
	gint i;
	gint xmin, xmax;
	GSList *iter;
	GtkXTreeViewColumn *column = NULL;
	
	/* TODO: add a stricter test */
	if (xtv->length == 0)
		return FALSE;

	for (i = 0; i <= xtv->last_row && xtv->y_pos[i] + xtv->y0 < y; i++) {
	}
	if (i > xtv->last_row)
		return FALSE;
	
	for (iter = xtv->columns; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		xmin = column->widget->allocation.x;
		xmax = xmin + column->widget->allocation.width;
		if (xmin <= x && x < xmax)
			break;
	}
	/* columns expands at full darea width, so we really should
	 * be in some column
	 */
	g_return_val_if_fail(iter, FALSE);

	if (expander) {
		*expander = FALSE;
		if ((column->col_id == xtv->expander_col_id) &&
		    (xtv->cell_prop[i] & GTK_XTREE_VIEW_HAS_CHILDREN)) {
			guint depth;
			depth = xtv->cell_prop[i] & GTK_XTREE_VIEW_DEPTH_MASK;
			xmin += depth * xtv->expander_width;
			xmax = xmin + xtv->expander_width + 1;
			if (xmin <= x && x <= xmax)
				*expander = TRUE;
		}
	}
		
	if (row)
		*row = i;
	if (col_id)
		*col_id = column->col_id;
	if (iterator)
		*iterator = xtv->iterators[i];
	return TRUE;
}

static gboolean darea_button_release_cb(GtkWidget *widget, 
		GdkEventButton *event,
		gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
	if (xtv->freeze)
		return TRUE;

	return FALSE;
}

void gtk_xtree_view_clear_selection(GtkXTreeView *xtv)
{
	int i;
	void *iterator;
	guint params;

	if (xtv->total_selected == 0 || xtv->length == 0)
		return;

	for (i = 0; i <= xtv->last_row; i++)
		xtv->cell_prop[i] &= ~GTK_XTREE_VIEW_SELECTED;
	
	/* begin searching for selected items at first displayed row */
	iterator = xtv->first_iterator;
	do {
		cache_get_row_params(xtv, iterator, &params);
		if (params & GTK_XTREE_VIEW_SELECTED) {
			params &= ~GTK_XTREE_VIEW_SELECTED;
			cache_set_row_params(xtv, iterator, params);
			xtv->total_selected--;
		}
	} while (xtv->total_selected && 
			xtv->data_iterator_func(&iterator, TRUE, FALSE, xtv->sort_col_id, xtv->sort_order, xtv->data_iterator_data));

	if (xtv->total_selected == 0)
		return;

	iterator = xtv->first_iterator;
	while (xtv->total_selected &&
			xtv->data_iterator_func(&iterator, FALSE, FALSE, xtv->sort_col_id, xtv->sort_order, xtv->data_iterator_data)) {
		cache_get_row_params(xtv, iterator, &params);
		if (params & GTK_XTREE_VIEW_SELECTED) {
			params &= ~GTK_XTREE_VIEW_SELECTED;
			cache_set_row_params(xtv, iterator, params);
			xtv->total_selected--;
		}
	}

	g_return_if_fail(xtv->total_selected == 0);
}

static gboolean darea_expose_event_cb(GtkWidget *widget, GdkEventExpose *event, gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
	gint i, col, w;
	gint depth;
	gint x, y, y0;
	gint has_children;
	gint is_selected;
	gint is_collapsed;
	gint first_c = 0;
	GSList *first_col = NULL;
	GSList *last_col = NULL;
	GSList *iter;
	gint first_row = -1;
	gint last_row = -1;
	GtkXTreeViewColumn *column;
	void **row_cells;
	GdkRectangle rect;
	GdkWindow *window = xtv->darea->window;
	GdkGC *black_gc = xtv->darea->style->black_gc;
	GdkGC *selected_gc = xtv->darea->style->base_gc[GTK_STATE_SELECTED];
	GdkGC **text_gc = xtv->darea->style->text_gc;
	GtkStateType state;
#ifdef DEBUG
	g_message("Exposed widget %p, darea %p\n", widget, xtv->darea);
#endif
	
#ifdef DEBUG
	g_message("Exposing x=%d y=%d w=%d h=%d\n", event->area.x, 
		event->area.y, event->area.width, event->area.height);
#endif
#if 0
	if (GTK_WIDGET_HAS_FOCUS(xtv->darea)) {
		GtkWidget *widget = GTK_WIDGET(xtv); 
		gtk_paint_focus(widget->style,
				widget->window,
				GTK_WIDGET_STATE(widget),
				NULL,
				widget,
				NULL,
				0, 0,
				widget->allocation.width,
				widget->allocation.height);
	}
#endif
	if (xtv->rows == 0)
		return TRUE;

#ifdef DEBUG 
	if (event->count)
		g_message("event->count %d\n", event->count);
#endif
	x = 0;
	y = 0;
	y0 = xtv->y0;

	for (col = 0, iter = xtv->columns; iter; iter = iter->next, col++) {
		column = (GtkXTreeViewColumn *) iter->data;
		first_col = iter;
		first_c = col;
		if (column->widget->allocation.x + column->widget->allocation.width >= event->area.x)
			
			break;
	}
	for (iter = first_col; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		last_col = iter;
		if (column->widget->allocation.x + column->widget->allocation.width > 
				event->area.x + event->area.width)
			break;
	}
	for (i = 0; i <= xtv->last_row; i++) {
		first_row = i;
		if (xtv->y_pos[i] + y0 >= event->area.y)
			break;
	}
	for (i = first_row; i <= xtv->last_row; i++) {
		last_row = i;
		if (xtv->y_pos[i] + y0 > event->area.y + event->area.height)
			break;
	}
	
	for (i = first_row; i <= last_row; i++) {
		is_selected = xtv->cell_prop[i] & GTK_XTREE_VIEW_SELECTED;
		has_children = xtv->cell_prop[i] & GTK_XTREE_VIEW_HAS_CHILDREN;
		is_collapsed = xtv->cell_prop[i] & GTK_XTREE_VIEW_COLLAPSED;
		depth = xtv->cell_prop[i] & GTK_XTREE_VIEW_DEPTH_MASK;
		y = (i > 0 ? xtv->y_pos[i - 1] : 0) + y0;
		rect.y = y;
		rect.height = xtv->y_pos[i] + y0 - y;
		state = (is_selected ? GTK_STATE_SELECTED : GTK_STATE_NORMAL);

		if (is_selected) {
			gdk_draw_rectangle(window,
					selected_gc, 
					TRUE,
					0, y,
					xtv->darea->allocation.width, rect.height);
		}
		if (xtv->iterators[i] == xtv->focused_iterator) {
			gtk_paint_focus(xtv->darea->style,
					window,
					GTK_STATE_NORMAL,
					NULL,
					xtv->darea,
					NULL,
					0, y, xtv->darea->allocation.width, rect.height);
		}


		row_cells = xtv->cells[i];

		col = first_c;
		for (iter = first_col;/*empty on purpose*/;iter = iter->next) {
			column = (GtkXTreeViewColumn *) iter->data;
			x = column->widget->allocation.x + 1;
			w = column->widget->allocation.width - 4;
			rect.x = x;
			rect.width = w;
			gdk_gc_set_clip_rectangle(text_gc[state], &rect);

			if (column->col_id == xtv->expander_col_id) {
				x += depth * xtv->expander_width;
				if (has_children) { 
					gtk_paint_arrow(xtv->darea->style,
							window,
							state,
							GTK_SHADOW_NONE,
							NULL,
							xtv->darea,
							"arrow",
							(is_collapsed ? 
							 GTK_ARROW_RIGHT :
							 GTK_ARROW_DOWN),
							FALSE,
							x, y + (rect.height - xtv->expander_width)/2 + 1,
							xtv->expander_width,
							xtv->expander_width);
							
				}
				x += xtv->expander_width + 2;
			}
			switch (column->type) {
				case GTK_XCOLUMN_TYPE_TEXT:
					gdk_draw_layout(window, 
							xtv->darea->style->text_gc[state],
							x, y + 1,
							(PangoLayout *) row_cells[col]);
					break;
				case GTK_XCOLUMN_TYPE_STOCKID:
				case GTK_XCOLUMN_TYPE_PIXBUF:
					if (row_cells[col]) {
					gdk_draw_pixbuf(window,
							black_gc,
							(GdkPixbuf *) row_cells[col],
							0, 0,
							x , y,
							-1, -1,
							GDK_RGB_DITHER_NORMAL, 0, 0);
					}
					break;
				default:
					g_assert_not_reached();
			}
			gdk_gc_set_clip_rectangle(black_gc, NULL);
			if (iter == last_col)
				break;
			col++;
		}
	}
	return TRUE;
}

static gboolean darea_fill(GtkXTreeView *xtv, void *first_iterator, guint first_row, guint last_row)
{
	void *iterator;
	void *last_iterator;
	gboolean has_children;
	guint depth;
	void *content;
	gchar *attribute;
	guint params;
	gboolean is_collapsed = 0;
	gboolean is_selected = 0;
	GSList *col_iter;
	guint row;
	guint col;
	GtkXTreeViewColumn *column;
	void **row_cells;
	gint height;
	gint h;
	gint y;
	gint old_y0;
	gint debug_i;
	GtkWidget *darea = xtv->darea;
	gboolean not_last;
	gboolean update_all = FALSE;

	if (xtv->length == 0 || xtv->freeze) 
		return TRUE;

	iterator = first_iterator;
	y = 0;
	row = first_row;
	
	if (row)
		y = xtv->y_pos[row - 1];
	debug_i = 0;
	do {
#ifdef DEBUG
	  g_message("row = %d, y = %d\n", row, y);
#endif
		if (row >= xtv->allocated_rows)
			add_row(xtv);

		debug_i++;

		xtv->row_params_func(iterator, 
				&has_children, 
				&depth, 
				xtv->row_params_data);

		cache_get_row_params(xtv, iterator, &params);
		params = (params & (GTK_XTREE_VIEW_SELECTED | GTK_XTREE_VIEW_COLLAPSED))
			| (has_children ? GTK_XTREE_VIEW_HAS_CHILDREN : 0) | depth;
		cache_set_row_params(xtv, iterator, params);
		is_selected = params & GTK_XTREE_VIEW_SELECTED;
		is_collapsed = params & GTK_XTREE_VIEW_COLLAPSED;

		row_cells = xtv->cells[row];
		xtv->iterators[row] = iterator;
		xtv->cell_prop[row] = params;

		height = 0;
		for (col = 0, col_iter = xtv->columns; 
				col_iter;
				col_iter = col_iter->next, col++) {
			column = (GtkXTreeViewColumn *) col_iter->data;
			xtv->row_data_func(iterator,
					column->col_id,
					&content,
					&attribute,
					xtv->row_data_data);
			switch (column->type) {
				case GTK_XCOLUMN_TYPE_TEXT:
					pango_layout_set_text
						((PangoLayout *) row_cells[col],
						 (content ? content : ""), -1);
					pango_layout_get_pixel_size
						((PangoLayout *) row_cells[col],
						 NULL, &h);
					break;
				case GTK_XCOLUMN_TYPE_PIXBUF:
					row_cells[col] = content;
					if (content) {
						h = gdk_pixbuf_get_height
							((GdkPixbuf *) content);
					} else {
						h = 0;
					}
					break;
				case GTK_XCOLUMN_TYPE_STOCKID:
					if (content) {
						row_cells[col] = 
							gtk_widget_render_icon
							(xtv->darea,
							 (char *) content,
							 GTK_ICON_SIZE_MENU,
							 NULL);
						h = gdk_pixbuf_get_height
							((GdkPixbuf *) row_cells[col]);
					} else {
						row_cells[col] = NULL;
						h = 0;
					}
					break;
				default:
					g_assert_not_reached();
			}
			h += 2;
			height = MAX(height, h);
		}
		xtv->y_pos[row] = y + height;
		y += height;
		row++;
#ifdef DEBUG
		g_message("row: %d last_row: %d y: %d height: %d\n",
		       row, last_row, y, darea->allocation.height);
#endif
		last_iterator = iterator;
		not_last = xtv->data_iterator_func(&iterator, 
				TRUE, 
				is_collapsed, 
				xtv->sort_col_id, 
				xtv->sort_order, 
				xtv->data_iterator_data);
	} while (row != last_row &&
			y < darea->allocation.height && 
			not_last);

	xtv->last_partial = ((last_row == 0 && y > darea->allocation.height) ?
			1 : 0);
	old_y0 = xtv->y0;
	xtv->y0 = 0;
	if (not_last == FALSE || last_row == 0) {
		xtv->last_row = row - 1;
		xtv->last_iterator = last_iterator;
		xtv->adjust_first = (not_last == FALSE &&
				y < darea->allocation.height);
		if (not_last == FALSE && y > darea->allocation.height) {
		/* Last data is partially viewed, scroll up to view it 
		 * completly.
		 */
			gint delta = y - darea->allocation.height;
			gint i, inc, t, s, j, y0;
			gint h = 0;
			void **row_cells_new, **row_cells_old;
			void *tmp;

			xtv->last_partial = 1;
			
			/* y0 is the offset of the first row */
			y0 = -delta;
			for (inc = 0; inc < xtv->last_row; inc++) {
				h += xtv->y_pos[inc];
				if (h > delta)
					break;
				y0 += xtv->y_pos[inc];
			}
			for (i = 0; i <= xtv->last_row - inc; i++) {
				xtv->y_pos[i] = 
					xtv->y_pos[i + inc] - delta - y0;
				xtv->iterators[i] = 
					xtv->iterators[i + inc];
				xtv->cell_prop[i] = 
					xtv->cell_prop[i + inc];
			}
			if (inc) {
				for (i = 0; i <= xtv->last_row - inc; i++) {
					t = (i) % (xtv->last_row + 1);
					s = (i + inc) % (xtv->last_row + 1);
					row_cells_new = xtv->cells[t];
					row_cells_old = xtv->cells[s];
					for (j = 0; j < xtv->col_nb; j++) {
						tmp = row_cells_new[j];
						row_cells_new[j] = row_cells_old[j];
						row_cells_old[j] = tmp;
					}
				}
				xtv->first_iterator = xtv->iterators[0];
				xtv->last_row -= inc;
				row -= inc;
				g_warning("We should update scrollbar\n");
			}
			xtv->y0 = y0;
		}
		if (xtv->rows != row) {
			xtv->rows = row;
			gtk_xtree_view_update_adjustment(xtv);
		}
	}
#ifdef DEBUG
	g_message("last_row = %d rows = %d\n", xtv->last_row, xtv->rows);
#endif
	update_all = old_y0 != xtv->y0;
	return update_all;
}

static void scroll_cb(GtkRange *range, gpointer axtv)
{
	GtkXTreeView *xtv = (GtkXTreeView *) axtv;
	gint increment;
	gboolean is_next;
	gint i;
	gint h, w;
	gdouble v = gtk_range_get_value(range);
	gint value = floor(v);
	GdkWindow *window = xtv->darea->window;
	GdkGC *black_gc = xtv->darea->style->black_gc;

	if (xtv->freeze || xtv->length == 0)
		return;
	
	gdk_window_freeze_updates(window);
	increment = value - xtv->first_value;

	xtv->first_value = value;

	if (increment == 0)
		return;
	
	is_next = (increment > 0);
	if (!is_next)
		increment = -increment;
	
	if (is_next && FALSE) {
		xtv->first_iterator = xtv->iterators[increment];
	} else {
		if (is_next) {
			guint params;
			for (i = 0; i < increment; i++) {
				cache_get_row_params(xtv, 
						xtv->first_iterator,
						&params);
				if (!xtv->data_iterator_func(&xtv->first_iterator,
							is_next,
							params & GTK_XTREE_VIEW_COLLAPSED,
							xtv->sort_col_id,
							xtv->sort_order,
							xtv->data_iterator_data))
				break;
			}
		} else {
			guint params = 0; /* first iterator is never hidden */
			for (i = 0; i < increment;) {
				/* If iterator is hidden we should request for
				 * parent
				 */
				if (!xtv->data_iterator_func(&xtv->first_iterator,
							FALSE,
							params & GTK_XTREE_VIEW_HIDDEN,
							xtv->sort_col_id,
							xtv->sort_order,
							xtv->data_iterator_data))
					break;
				cache_get_row_params(xtv, xtv->first_iterator,
						&params);
				if ((params & GTK_XTREE_VIEW_HIDDEN) == 0)
					i++;
			}
		}
	}

	w = xtv->darea->allocation.width;
	h = xtv->darea->allocation.height;
	if (increment >= xtv->rows) {
		darea_fill(xtv, xtv->first_iterator, 0, 0);
		gtk_widget_queue_draw(xtv->darea);
	} else {
		if (is_next) {
			GdkRectangle rect;
			gint j, s, t;
			gint *y_pos = xtv->y_pos;
			void **iterators = xtv->iterators;
			void **row_cells_new;
			void **row_cells_old;
			void *tmp;
			gint last_row = xtv->last_row;
			gint delta = y_pos[increment - 1];
			gint height = y_pos[last_row - 1] - delta;

			height = y_pos[last_row - 1] - delta;
 			gdk_draw_drawable(window, 
					black_gc, 
					window,
					0, delta,
					0, 0,
					w, height);

			for (i = 0; i < last_row - increment; i++) {
				y_pos[i] = y_pos[i + increment] - delta;
				iterators[i] = iterators[i + increment];
				xtv->cell_prop[i] =
					xtv->cell_prop[i + increment];
			}
			for (i = 0; i <= last_row - increment; i++) {
				t = (i) % (last_row + 1);
				s = (i + increment) % (last_row + 1);
				row_cells_new = xtv->cells[t];
				row_cells_old = xtv->cells[s];
				for (j = 0; j < xtv->col_nb; j++) {
					tmp = row_cells_new[j];
					row_cells_new[j] = row_cells_old[j];
					row_cells_old[j] = tmp;
				}
			}
			/* Should check children if any */
			rect.x = 0; rect.y = height; rect.width = w; rect.height = h - height;
			if (darea_fill(xtv, xtv->last_iterator,
						xtv->last_row - increment, 0)) {
				rect.y = 0; rect.height = h;
			}
			gdk_window_invalidate_rect(window, &rect,
					FALSE);
			
		} else {
			GdkRectangle rect;
			gint i, j, s, t;
			gint height;
			gint delta;
			void **row_cells_new;
			void **row_cells_old;
			void *tmp;
			gboolean update_all;

			for (i = 0; i <= xtv->last_row; i++) {
				xtv->y_pos_tmp[i] = xtv->y_pos[i];
				xtv->iterators_tmp[i] = xtv->iterators[i];
				xtv->cell_prop_tmp[i] = xtv->cell_prop[i];
			}
			for (i = xtv->last_row; i >= increment; i--) {
				t = (i - increment) % (xtv->last_row + 1);
				s = (i) % (xtv->last_row + 1);
				row_cells_new = xtv->cells[t];
				row_cells_old = xtv->cells[s];
				for (j = 0; j < xtv->col_nb; j++) {
					tmp = row_cells_new[j];
					row_cells_new[j] = row_cells_old[j];
					row_cells_old[j] = tmp;
				}
			}
			update_all = darea_fill(xtv, xtv->first_iterator, 0,
					increment);
			delta = xtv->y_pos[increment - 1];
			for (i = 0; i <= xtv->last_row; i++) {
				if (i + increment == xtv->allocated_rows)
					add_row(xtv);
				xtv->y_pos[i + increment] = 
					xtv->y_pos_tmp[i] + delta;
				xtv->iterators[i + increment] = 
					xtv->iterators_tmp[i];
				xtv->cell_prop[i + increment] =
					xtv->cell_prop_tmp[i];
				xtv->last_row = i + increment;
				if (xtv->y_pos[i + increment] >= h)
					break;
			}
			xtv->last_iterator = xtv->iterators[xtv->last_row];
			if (xtv->last_row + 1 > xtv->rows) {
#ifdef DEBUG
				g_message("scrolling up changed number of displayed rows from %d to %d\n", xtv->rows, xtv->last_row + 1);
#endif
				xtv->rows = xtv->last_row + 1;
			}
			height = h - delta;
			gdk_draw_drawable(window,
					black_gc, 
					window,
					0, 0, 
					0, delta, 
					w, height);
			rect.x = 0; rect.y = 0; rect.width = w; rect.height = (update_all ? h : delta);
			gdk_window_invalidate_rect(window, &rect, FALSE);
		}
	}
	gdk_window_thaw_updates(window);
	gdk_window_process_updates(window, FALSE);
}

static void cache_set_row_params(GtkXTreeView *xtv, void *iterator, guint params)
{
	guint value = params | GTK_XTREE_VIEW_NON_NULL;
	g_hash_table_insert(xtv->cache, iterator, (gpointer) value);
}

static void cache_get_row_params(GtkXTreeView *xtv, void *iterator, guint *params)
{
	guint value;

	/* GLib hash table store pointers which are non NULL
	 * we instead want to store the value instead of the pointer,
	 * the value being possibly 0, we make sure all stored values
	 * are non null by setting a single bit
	 */

        value = (guint) g_hash_table_lookup(xtv->cache, iterator);
	if (value == 0) {
		value = (xtv->default_collapsed ? GTK_XTREE_VIEW_COLLAPSED : 0);
		value |= GTK_XTREE_VIEW_NON_NULL;
		g_hash_table_insert(xtv->cache, iterator, (gpointer) value);
	}
	*params = value & ~GTK_XTREE_VIEW_NON_NULL;
}

static void add_row(GtkXTreeView *xtv)
{
	GSList *iter;
	void **row_cells;
	GtkXTreeViewColumn *column;
	gint col;
	
	xtv->allocated_rows++;

	xtv->cells = g_renew(void **, xtv->cells, xtv->allocated_rows);
	xtv->cell_prop = g_renew(gint, xtv->cell_prop, xtv->allocated_rows);
	xtv->cell_prop_tmp = g_renew(gint, xtv->cell_prop_tmp, xtv->allocated_rows);
	xtv->y_pos = g_renew(gint, xtv->y_pos, xtv->allocated_rows);
	xtv->y_pos_tmp = g_renew(gint, xtv->y_pos_tmp, xtv->allocated_rows);
	xtv->iterators = g_renew(void *, xtv->iterators, xtv->allocated_rows);
	xtv->iterators_tmp = g_renew(void *, xtv->iterators_tmp, xtv->allocated_rows);

	xtv->cells[xtv->allocated_rows - 1] = row_cells = g_new(void *, g_slist_length(xtv->columns));
	
	for (col = 0, iter = xtv->columns; iter; iter = iter->next, col++) {
		column = (GtkXTreeViewColumn *) iter->data;
		switch(column->type) {
			case GTK_XCOLUMN_TYPE_TEXT:
				row_cells[col] = gtk_widget_create_pango_layout(xtv->darea, NULL);
				break;
			default:
				row_cells[col] = NULL;
		}
	}
}

GtkWidget *gtk_xtree_view_new(void)
{
	return GTK_WIDGET(g_object_new(gtk_xtree_view_get_type(), NULL));
}

gboolean gtk_xtree_view_insert_column(GtkXTreeView *xtv, 
		guint col, 
		guint col_id, 
		GtkXColumnType type, 
		gboolean is_sortable, 
		GtkSortType sort_order, 
		GtkWidget *title_widget, 
		gchar *title_string)
{
	GtkXTreeViewColumn *column;
	gboolean fill;

	g_return_val_if_fail(col <= g_slist_length(xtv->columns), FALSE);
	g_return_val_if_fail(col_id > 0, FALSE);

	column = g_new(GtkXTreeViewColumn, 1);

	column->col_id = col_id;
	column->type = type;
	column->is_sortable = is_sortable;
	column->sort_order = sort_order;
	column->treeview = xtv;
	
	create_title_button(column, title_widget, title_string);
	if (xtv->expansible_col_id == 0 || xtv->expansible_col_id == col_id)
		fill = TRUE;
	else
		fill = FALSE;
	gtk_box_pack_start(GTK_BOX(xtv->hbox), column->widget, fill, TRUE, 0);

	xtv->columns = g_slist_append(xtv->columns, column);
	xtv->col_nb = g_slist_length(xtv->columns);
	return TRUE;
}

static void create_title_button(GtkXTreeViewColumn *column, 
		GtkWidget *title_widget, 
		gchar *title_string)
{
	GtkWidget *button;
	GtkStyle *style;
	GtkWidget *hbox;
	GtkWidget *label;
	GtkWidget *arrow;


	button = gtk_button_new();
	style = gtk_rc_get_style_by_paths(gtk_widget_get_settings(button), NULL,
			"*.GtkTreeView.GtkButton", G_OBJECT_TYPE(button));
	gtk_widget_set_style(button, style);

	hbox = gtk_hbox_new(FALSE, 0);
	
	if (title_widget) {
		gtk_box_pack_start(GTK_BOX(hbox), title_widget, FALSE, FALSE, 0);
		gtk_widget_show(title_widget);
	}

	label = gtk_label_new_with_mnemonic(title_string);
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0);
	gtk_widget_show(label);

	arrow = gtk_arrow_new(GTK_ARROW_UP, GTK_SHADOW_IN);
	gtk_box_pack_end(GTK_BOX(hbox), arrow, FALSE, FALSE, 0);

	gtk_container_add(GTK_CONTAINER(button), hbox);
	gtk_widget_show(hbox);

	gtk_widget_show(button);

	column->widget = button;
	column->hbox = hbox;
	column->user_widget = title_widget;
	column->label = label;
	column->arrow = arrow;

	/* Header signals */
	g_signal_connect(G_OBJECT(button), "realize", 
			G_CALLBACK(header_realize_cb), column);
	g_signal_connect(G_OBJECT(column->treeview->hbox), "button-press-event", 
			G_CALLBACK(header_button_press_cb),
			column);
	g_signal_connect(G_OBJECT(button), "button-release-event", 
			G_CALLBACK(header_button_release_cb),
			column);
	g_signal_connect(G_OBJECT(button), "clicked", 
			G_CALLBACK(header_clicked_cb),
			column);
	g_signal_connect(G_OBJECT(button), "event",
			G_CALLBACK(header_event_cb),
			column);
	g_signal_connect(G_OBJECT(button), "scroll-event",
			G_CALLBACK(header_scroll_event_cb),
			column);
}

static void header_realize_cb(GtkWidget *button, gpointer acolumn)
{
	GdkWindow *resize_grip;
	GdkWindowAttr attr;
	guint attributes_mask;
	GtkXTreeViewColumn *column = (GtkXTreeViewColumn *) acolumn;

	attr.window_type = GDK_WINDOW_CHILD;
	attr.wclass = GDK_INPUT_ONLY;
	attr.visual = gtk_widget_get_visual (button);
	attr.colormap = gtk_widget_get_colormap (button);
	attr.event_mask = gtk_widget_get_events (button) |
		(GDK_BUTTON_PRESS_MASK |
		 GDK_BUTTON_RELEASE_MASK |
		 GDK_POINTER_MOTION_MASK |
		 GDK_POINTER_MOTION_HINT_MASK |
		 GDK_KEY_PRESS_MASK);
	attributes_mask = GDK_WA_CURSOR | GDK_WA_X | GDK_WA_Y;
	attr.cursor = gdk_cursor_new_for_display (gtk_widget_get_display (button),
			GDK_SB_H_DOUBLE_ARROW);
	attr.y = 0;
	attr.width = 3;
	attr.height = button->allocation.height;

	attr.x = (button->allocation.x + button->allocation.width) - 3;

	resize_grip = gdk_window_new (button->window,
			&attr, attributes_mask);
	gdk_cursor_unref(attr.cursor);
	gdk_window_show(resize_grip);
	gdk_window_raise(resize_grip);

	column->resize_grip = resize_grip;
	column->being_resized = FALSE;
}

static gboolean header_button_release_cb(GtkWidget *widget, GdkEventButton *event, gpointer acolumn)
{
	GtkXTreeViewColumn *column = (GtkXTreeViewColumn *) acolumn;
	column->being_resized = FALSE;

	return (event->button == 1 ? FALSE : TRUE);
}

static gboolean header_scroll_event_cb(GtkWidget *widget, 
		GdkEvent *event, gpointer acolumn)
{
	GtkXTreeViewColumn *column = (GtkXTreeViewColumn *) acolumn;
	GtkXTreeView *xtv = column->treeview;
	GdkEventScroll *scroll = (GdkEventScroll *) event;
	GtkSortType sort_order;

	if (scroll->direction == GDK_SCROLL_DOWN)
		sort_order = GTK_SORT_ASCENDING;
	else if (scroll->direction == GDK_SCROLL_UP)
		sort_order = GTK_SORT_DESCENDING;
	else
		return FALSE;
	
	gtk_xtree_view_set_sort_column(xtv, column->col_id, sort_order);
	/* Showing/hiding the arrows may change column width,
	 * so we redraw things.
	 */
	gtk_widget_queue_draw(xtv->darea);

	return FALSE;
}

static void header_clicked_cb(GtkButton *button, gpointer acolumn)
{
	GtkXTreeViewColumn *column = (GtkXTreeViewColumn *) acolumn;
	GtkXTreeView *xtv = column->treeview;
	GtkSortType sort_order;

	if (xtv->sort_col_id == column->col_id) {
		sort_order = (column->sort_order == GTK_SORT_ASCENDING ?
				GTK_SORT_DESCENDING : GTK_SORT_ASCENDING);
	} else {
		sort_order = column->sort_order;
	}
	gtk_xtree_view_set_sort_column(xtv, column->col_id, sort_order);
	/* Showing/hiding the arrows may change column width,
	 * so we redraw things.
	 */
	gtk_widget_queue_draw(xtv->darea);
}
	
static gboolean header_button_press_cb(GtkWidget *widget, 
		GdkEventButton *event, gpointer acolumn)
{
	GtkXTreeViewColumn *column = (GtkXTreeViewColumn *) acolumn;
	if (column->treeview->freeze)
		return FALSE;
	if (event->window == column->resize_grip) {
		if (gdk_pointer_grab(column->resize_grip, FALSE,
					GDK_POINTER_MOTION_HINT_MASK |
					GDK_BUTTON1_MOTION_MASK |
					GDK_BUTTON_RELEASE_MASK,
					NULL, NULL, event->time))
			return FALSE;
		column->being_resized = TRUE;
		return TRUE;
	}
	return FALSE;
	
}

static gboolean header_event_cb(GtkWidget *widget, 
		GdkEvent *event,
		gpointer acolumn)
{
	GtkXTreeViewColumn *column = (GtkXTreeViewColumn *) acolumn;

	g_return_val_if_fail(event != NULL, FALSE);

	if (event->type == GDK_MOTION_NOTIFY)
		g_message(column->being_resized ? "resizin" :"moving...\n");
	return FALSE;
}

gboolean gtk_xtree_view_set_sort_column(GtkXTreeView *xtv, guint col_id, GtkSortType sort_order)
{
	GSList *iter;
	GtkXTreeViewColumn *column;
	
	if (xtv->sort_order == sort_order && xtv->sort_col_id == col_id)
		return FALSE;

	xtv->sort_col_id = 0;
	for (iter = xtv->columns; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		if (column->col_id == col_id) {
			column->sort_order = sort_order;
			xtv->sort_order = sort_order;
			xtv->sort_col_id = col_id;
			gtk_widget_show(column->arrow);
			gtk_arrow_set(GTK_ARROW(column->arrow), 
					(sort_order == GTK_SORT_ASCENDING ?
					 GTK_ARROW_DOWN : 
					 GTK_ARROW_UP),
					GTK_SHADOW_NONE);
		} else {
			gtk_widget_hide(column->arrow);
		}
	}
	g_signal_emit(G_OBJECT(xtv),
			gtkxtreeview_signals[SORT_CHANGED_SIGNAL],
			0);
	return TRUE;
}
		
void gtk_xtree_view_set_expander_column(GtkXTreeView *xtv, guint col_id)
{
	GSList *iter;
	GtkXTreeViewColumn *column;
	
	xtv->expander_col_id = 0;
	for (iter = xtv->columns; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		if (column->col_id == col_id) {
			xtv->expander_col_id = col_id;
		}
	}
}

void gtk_xtree_view_set_column_width(GtkXTreeView *xtv, guint col_id, guint width)
{	
	GSList *iter;
	GtkXTreeViewColumn *column;
	
	for (iter = xtv->columns; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		if (column->col_id == col_id) {
			gtk_widget_set_size_request(column->widget, width, -1);
		}
	}

}

void gtk_xtree_view_set_expansible_column(GtkXTreeView *xtv, guint col_id)
{
	GSList *iter;
	GtkXTreeViewColumn *column;
	gboolean fill;

	xtv->expansible_col_id = 0;
	for (iter = xtv->columns; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		fill = FALSE;
		if (column->col_id == col_id) {
			xtv->expansible_col_id = col_id;
			fill = TRUE;
		} 
		gtk_box_set_child_packing(GTK_BOX(xtv->hbox),
				column->widget, fill, TRUE, 0, GTK_PACK_START);
		
	}
}

void gtk_xtree_view_set_row_data_func(GtkXTreeView *xtv, GtkXTreeViewRowDataFunc func, gpointer data)
{
	xtv->row_data_func = func;
	xtv->row_data_data = data;
}
void gtk_xtree_view_set_row_params_func(GtkXTreeView *xtv, GtkXTreeViewRowParamsFunc func, gpointer data)
{
	xtv->row_params_func = func;
	xtv->row_params_data = data;
}

void gtk_xtree_view_set_data_iterator_func(GtkXTreeView *xtv, GtkXTreeViewDataIteratorFunc func, gpointer data)
{
	xtv->data_iterator_func = func;
	xtv->data_iterator_data = data;
}

void gtk_xtree_view_set_first_row(GtkXTreeView *xtv, void *iterator)
{
	xtv->first_iterator = iterator;
	xtv->first_value = 0;
	xtv->focused_iterator = iterator;
	xtv->focused_value = 0;
	gtk_range_set_value(GTK_RANGE(xtv->vscrollbar), 0.0);
}

void gtk_xtree_view_set_match_func(GtkXTreeView *xtv, GtkXTreeViewMatchFunc func, gpointer data)
{
	xtv->match_func = (func ? func : default_match_func);
	xtv->match_data = data;
}



void gtk_xtree_view_freeze(GtkXTreeView *xtv)
{
	xtv->freeze++;
}

void gtk_xtree_view_thaw(GtkXTreeView *xtv)
{
	if (xtv->freeze)
		xtv->freeze--;
	if (xtv->freeze == 0)
		gtk_xtree_view_update(xtv);
}

void gtk_xtree_view_reset(GtkXTreeView *xtv)
{
	xtv->freeze = 0;
	xtv->length = 0;
	xtv->displayed = 0;
	xtv->rows = 0;
	xtv->first_iterator = NULL;
	xtv->first_value = 0;
	xtv->focused_iterator = NULL;
	xtv->focused_value = 0;
	xtv->total_selected = 0;
	g_hash_table_destroy(xtv->cache);
	xtv->cache = g_hash_table_new(NULL, NULL);
	gtk_xtree_view_update_adjustment(xtv);
}

void gtk_xtree_view_set_length(GtkXTreeView *xtv, guint length)
{
	/* TODO: we should invalidate the display */
	xtv->length = length;
	if (length == 0) {
		gtk_xtree_view_reset(xtv);
	}
	xtv->displayed = length;
	gtk_xtree_view_update_adjustment(xtv);
}

static void gtk_xtree_view_update_adjustment(GtkXTreeView *xtv)
{
	GtkAdjustment *adj;

	adj = gtk_range_get_adjustment(GTK_RANGE(xtv->vscrollbar));

	adj->lower = 0.0;
	adj->upper = (xtv->displayed > 0 ? xtv->displayed : 1.0);
	adj->step_increment = 1.0;
	adj->page_increment = (xtv->rows > 0 ? xtv->rows : 1.0);
	adj->page_size = (xtv->rows > 0 ? xtv->rows : 1.0);
	gtk_adjustment_changed(adj);
	if (adj->value > xtv->displayed - xtv->rows) {
		adj->value = xtv->displayed - xtv->rows;
		gtk_adjustment_value_changed(adj);
	}
}

void gtk_xtree_expand_all(GtkXTreeView *xtv)
{
}

static void update_resize_grips(GtkXTreeView *xtv)
{
	GSList *iter;
	GtkXTreeViewColumn *column;

	for (iter = xtv->columns; iter; iter = iter->next) {
		column = (GtkXTreeViewColumn *) iter->data;
		if (GTK_WIDGET_REALIZED(column->widget)) {
			gdk_window_move(column->resize_grip,
					column->widget->allocation.x +
					column->widget->allocation.width - 3,
					0);
			gdk_window_show(column->resize_grip);
			gdk_window_raise(column->resize_grip);
		}
	}
}

void gtk_xtree_view_update(GtkXTreeView *xtv)
{
	GdkRectangle rect;
	if (GTK_WIDGET_REALIZED(xtv->darea)) {
		darea_fill(xtv, xtv->first_iterator, 0, 0);
		rect.x = 0;
		rect.y = 0;
		rect.width = xtv->darea->allocation.width;
		rect.height = xtv->darea->allocation.height;
		gdk_window_invalidate_rect(xtv->darea->window, &rect, FALSE);
		gdk_window_process_updates(xtv->darea->window, FALSE);
	}
}

void *gtk_xtree_view_get_focused(GtkXTreeView *xtv)
{
	return xtv->focused_iterator;
}

uint32_t gtk_xtree_view_select_all(GtkXTreeView *xtv, void ***selection)
{
	uint32_t i = xtv->first_value;
	void *iterator = xtv->first_iterator;
	guint params;
	void **sel = NULL;

	if (xtv->length == 0)
		return 0;
	/* When we select all, we select all displayed items. Children
	 * of collapsed items are not selected
	 */
	if (selection)
		sel = g_new(void *, xtv->displayed);
	do {
		cache_get_row_params(xtv, iterator, &params);
		params |= GTK_XTREE_VIEW_SELECTED;
		cache_set_row_params(xtv, iterator, params);
		if (sel)
			sel[i] = iterator;
		i++;
	} while (i < xtv->displayed && xtv->data_iterator_func(&iterator,
				TRUE,
				FALSE,
				xtv->sort_col_id,
				xtv->sort_order,
				xtv->data_iterator_data));
	i = xtv->first_value;

	iterator = xtv->first_iterator;
				
	while (i && xtv->data_iterator_func(&iterator, FALSE, FALSE,
				xtv->sort_col_id,
				xtv->sort_order,
				xtv->data_iterator_data)) {
		cache_get_row_params(xtv, iterator, &params);
		params |= GTK_XTREE_VIEW_SELECTED;
		cache_set_row_params(xtv, iterator, params);
		i--;
		if (sel)
			sel[i] = iterator;
		
	}
	
	if (i != 0) {
		g_error("could not find correct number of total items\n");
	}
	if (sel)
		*selection = sel;
	xtv->total_selected = xtv->displayed;
	gtk_xtree_view_update(xtv);
	return xtv->total_selected;
}

uint32_t gtk_xtree_view_invert_selection(GtkXTreeView *xtv, void ***selection)
{
	uint32_t i = 0;
	uint32_t new_total_selected = xtv->displayed - xtv->total_selected;
	void *iterator = xtv->first_iterator;
	guint params;
	void **sel = NULL;

	if (xtv->length == 0)
		return 0;
	
	if (selection && new_total_selected)
		sel = g_new(void *, new_total_selected);
	do {
		cache_get_row_params(xtv, iterator, &params);
		params ^= GTK_XTREE_VIEW_SELECTED;
		cache_set_row_params(xtv, iterator, params);
		if (params & GTK_XTREE_VIEW_SELECTED) {
			if (sel)
				sel[i] = iterator;
			i++;
		}
	} while (xtv->data_iterator_func(&iterator,
				TRUE,
				FALSE,
				xtv->sort_col_id,
				xtv->sort_order,
				xtv->data_iterator_data));
	if (sel && i < new_total_selected) {
		uint32_t j;
		for (j = 0; j < i; j++) {
			sel[new_total_selected - i + j] = sel[j];
		}
	}

	i = new_total_selected - i;

	iterator = xtv->first_iterator;
				
	while (xtv->data_iterator_func(&iterator, FALSE, FALSE,
				xtv->sort_col_id,
				xtv->sort_order,
				xtv->data_iterator_data)) {
		cache_get_row_params(xtv, iterator, &params);
		params ^= GTK_XTREE_VIEW_SELECTED;
		cache_set_row_params(xtv, iterator, params);
		if (params & GTK_XTREE_VIEW_SELECTED) {
			i--;
			if (sel)
				sel[i] = iterator;
		}
	}
	
	if (i != 0) {
		g_error("could not find correct number of total items\n");
	}
	if (sel)
		*selection = sel;
	xtv->total_selected = new_total_selected;
	gtk_xtree_view_update(xtv);
	return xtv->total_selected;
}

uint32_t gtk_xtree_view_get_selection(GtkXTreeView *xtv, 
		void ***selection)
{
	if (selection && xtv->total_selected) {
		void *iterator = xtv->first_iterator;
		void **sel = NULL;
		uint32_t i = 0;
		guint params;
		
		sel = g_new(void *, xtv->total_selected);
		
		do {
			cache_get_row_params(xtv, iterator, &params);
			if (params & GTK_XTREE_VIEW_SELECTED) {
				sel[i] = iterator;
				i++;
			}
		} while (i < xtv->total_selected &&
				xtv->data_iterator_func(&iterator,
					TRUE,
					FALSE,
					xtv->sort_col_id,
					xtv->sort_order,
					xtv->data_iterator_data));

		if (i < xtv->total_selected) {
			uint32_t j;
			for (j = 0; j < i; j++) {
				sel[xtv->total_selected - i + j] = sel[j];
			}
			j = xtv->total_selected - i;
			iterator = xtv->first_iterator;
			while (xtv->data_iterator_func(&iterator, FALSE, FALSE,
						xtv->sort_col_id,
						xtv->sort_order,
						xtv->data_iterator_data)
					&& j > 0) {
				cache_get_row_params(xtv, iterator, &params);
				if (params & GTK_XTREE_VIEW_SELECTED) {
					j--;
					sel[j] = iterator;
				}
			}
			if (j != 0) {
				g_error("could not found the correct number of selected items\n");
			}
		}
		*selection = sel;
	}

	return xtv->total_selected;
}

gboolean gtk_xtree_view_next_match(GtkXTreeView *xtv, gboolean activate_match)
{
	void *iterator;
	uint32_t index;
	gboolean found;

	if (xtv->freeze)
		return FALSE;

	if (xtv->focused_iterator) {
		iterator = xtv->focused_iterator;
		index = xtv->focused_value;
		found = FALSE;
	} else {
		iterator = xtv->first_iterator;
		index = xtv->first_value;
		found = xtv->match_func(iterator, xtv->match_data);
	}

	while (!found && xtv->data_iterator_func(&iterator, 
						 TRUE, 
						 FALSE, 
						 xtv->sort_col_id, 
						 xtv->sort_order, 
						 xtv->data_iterator_data))
	  {
	    found = xtv->match_func(iterator, xtv->match_data);
	    index++;
	  }

	if (found) {
		GtkAdjustment *adj = 
			gtk_range_get_adjustment(GTK_RANGE(xtv->vscrollbar));
		xtv->focused_iterator = iterator;
		xtv->focused_value = index;
		if (activate_match) {
			gtk_xtree_view_clear_selection(xtv);
			select_iterator(xtv, iterator, FALSE);
			g_signal_emit(G_OBJECT(xtv),
					gtkxtreeview_signals[ROW_ACTIVATED_SIGNAL],	
					0,
					xtv->focused_iterator,
					0);
		}
		gtk_xtree_view_update(xtv);
		if (index >= xtv->first_value + xtv->last_row) {
			gtk_range_set_value(GTK_RANGE(xtv->vscrollbar),
					index - adj->page_increment * 0.5 + 1);
		}
	}
	return found;
}

gboolean gtk_xtree_view_prev_match(GtkXTreeView *xtv, gboolean activate_match)
{
	void *iterator;
	uint32_t index;
	gboolean found = FALSE;

	if (xtv->freeze)
		return FALSE;

	if (xtv->focused_iterator) {
		iterator = xtv->focused_iterator;
		index = xtv->focused_value;
	} else {
	  /* FIXME: if no focused iterator, search from bottom*/
		iterator = xtv->first_iterator;
		index = xtv->first_value;
	}
	if (index == 0)
		return FALSE;
	index--;
	while (xtv->data_iterator_func(&iterator, 
				FALSE, 
				FALSE, 
				xtv->sort_col_id, 
				xtv->sort_order, 
				xtv->data_iterator_data) && 
			!(found = xtv->match_func(iterator, xtv->match_data))) {
		index--;
	}

	if (found) {
		GtkAdjustment *adj = 
			gtk_range_get_adjustment(GTK_RANGE(xtv->vscrollbar));
		xtv->focused_iterator = iterator;
		xtv->focused_value = index;

		if (activate_match) {
			gtk_xtree_view_clear_selection(xtv);
			select_iterator(xtv, iterator, FALSE);
			g_signal_emit(G_OBJECT(xtv),
					gtkxtreeview_signals[ROW_ACTIVATED_SIGNAL],	
					0,
					xtv->focused_iterator,
					0);
		}
		gtk_xtree_view_update(xtv);
		if (index < xtv->first_value) {
			gtk_range_set_value(GTK_RANGE(xtv->vscrollbar),
					index - adj->page_increment * 0.5 + 1);
		}
	}
	return found;
}

uint32_t gtk_xtree_view_select_all_matches(GtkXTreeView *xtv, gboolean new_selection)
{
	void *iterator;
	uint32_t count = 0;

	if (xtv->freeze)
		return FALSE;
	if (new_selection) {
		gtk_xtree_view_clear_selection(xtv);
	}

	iterator = xtv->first_iterator;
	do {
		if (xtv->match_func(iterator, xtv->match_data)) {
			select_iterator(xtv, iterator, FALSE);
			count++;

		}
	} while (xtv->data_iterator_func(&iterator, 
				TRUE, 
				FALSE, 
				xtv->sort_col_id, 
				xtv->sort_order, 
				xtv->data_iterator_data));

	iterator = xtv->first_iterator;
	while (xtv->data_iterator_func(&iterator, 
				FALSE, 
				FALSE, 
				xtv->sort_col_id, 
				xtv->sort_order, 
				xtv->data_iterator_data)) {
		if (xtv->match_func(iterator, xtv->match_data)) {
			select_iterator(xtv, iterator, FALSE);
			count++;
		}
	}

	if (count) {
		gtk_xtree_view_update(xtv);
	}
	return count;
}

static gboolean default_match_func(void *iterator, gpointer data)
{
	return TRUE;
}
		

gboolean gtk_xtree_view_toggle_node(GtkXTreeView *xtv, void *iterator)
{
	guint params;

	cache_get_row_params(xtv, iterator, &params);

	if (!(params & GTK_XTREE_VIEW_HAS_CHILDREN))
		return FALSE;

	if (params & GTK_XTREE_VIEW_COLLAPSED)
		return expand_node(xtv, iterator, params);
	else
		return collapse_node(xtv, iterator, params);
}

gboolean gtk_xtree_view_collapse_node(GtkXTreeView *xtv, void *iterator)
{
	guint params;

	cache_get_row_params(xtv, iterator, &params);

	if (!(params & GTK_XTREE_VIEW_HAS_CHILDREN))
		return FALSE;
	if (params & GTK_XTREE_VIEW_COLLAPSED)
		return FALSE;
	return collapse_node(xtv, iterator, params);
}

gboolean gtk_xtree_view_expand_node(GtkXTreeView *xtv, void *iterator)
{
	guint params;

	cache_get_row_params(xtv, iterator, &params);

	if (!(params & GTK_XTREE_VIEW_HAS_CHILDREN))
		return FALSE;
	if (params & GTK_XTREE_VIEW_COLLAPSED)
		return FALSE;
	return expand_node(xtv, iterator, params);
}

static gboolean expand_node(GtkXTreeView *xtv, void *iterator, guint params)
{
	const guint depth = params & GTK_XTREE_VIEW_DEPTH_MASK;
	void *next_iterator;
	guint next_params;
	guint next_depth;
	gboolean is_collapsed;

	params &= ~GTK_XTREE_VIEW_COLLAPSED;
	cache_set_row_params(xtv, iterator, params);

	next_iterator = iterator;
	is_collapsed = FALSE;

	while (xtv->data_iterator_func(&next_iterator, TRUE, is_collapsed, 
				xtv->sort_col_id,
				xtv->sort_order,
				xtv->data_iterator_data)) {
		cache_get_row_params(xtv, next_iterator, &next_params);
		next_depth = next_params & GTK_XTREE_VIEW_DEPTH_MASK;
		is_collapsed = next_params & GTK_XTREE_VIEW_COLLAPSED;
		if (next_depth <= depth)
			break;
		else {
			next_params &= ~GTK_XTREE_VIEW_HIDDEN;
			cache_set_row_params(xtv, next_iterator, next_params);
			xtv->displayed++;
		}
	}
	gtk_xtree_view_update_adjustment(xtv);
	gtk_xtree_view_update(xtv);
	return TRUE;
}

static gboolean collapse_node(GtkXTreeView *xtv, void *iterator, guint params)
{
	const guint depth = params & GTK_XTREE_VIEW_DEPTH_MASK;
	void *next_iterator;
	guint next_params;
	guint next_depth;
	gboolean is_collapsed;

	params |= GTK_XTREE_VIEW_COLLAPSED;
	cache_set_row_params(xtv, iterator, params);

	next_iterator = iterator;
	is_collapsed = FALSE;

	while (xtv->data_iterator_func(&next_iterator, TRUE, is_collapsed, 
				xtv->sort_col_id,
				xtv->sort_order,
				xtv->data_iterator_data)) {
		cache_get_row_params(xtv, next_iterator, &next_params);
		next_depth = next_params & GTK_XTREE_VIEW_DEPTH_MASK;
		is_collapsed = next_params & GTK_XTREE_VIEW_COLLAPSED;
		if (next_depth <= depth)
			break;
		else {
			next_params |= GTK_XTREE_VIEW_HIDDEN;
			cache_set_row_params(xtv, next_iterator, next_params);
			xtv->displayed--;
		}
	}
	gtk_xtree_view_update_adjustment(xtv);
	gtk_xtree_view_update(xtv);
	return TRUE;
}
#endif
