/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_PYTHON

#include <Python.h>

#include "xetpan.h"
#include "ui.h"

#include "xep-python.h"

/* FIXME: clean up: do not do work in python module */
#include <libetpan/maildriver_types.h>
#include "message.h"

#ifdef USE_PYTHON_WITH_THREAD_SUPPORT
static PyThreadState *main_thread_state;
#endif

/* Static functions declarations */

static PyObject *
XePPython_get_folder_list(PyObject *self, PyObject *args);

static PyObject *
XePPython_rule_new(PyObject *self, PyObject *args);

static PyObject *
XePPython_rule_get(PyObject *self, PyObject *args);

static PyObject *
XePPython_condition_new(PyObject *self, PyObject *args);

static PyObject *
XePPython_condition_get(PyObject *self, PyObject *args);

static PyObject *
XePPython_action_new(PyObject *self, PyObject *args);

static PyObject *
XePPython_action_get(PyObject *self, PyObject *args);

static PyObject *
XePPython_folder_get_rules(PyObject *self, PyObject *args);

static PyObject *
XePPython_folder_add_rule(PyObject *self, PyObject *args);

static PyObject *
XePPython_folder_remove_rule(PyObject *self, PyObject *args);

#if 0
static PyObject *
XePPython_get_mainview_list(PyObject *self, PyObject *args);
#endif

static PyObject *
XePPython_update_folder(PyObject *self, PyObject *args);

static PyObject *
XePPython_folder_get_uids(PyObject *self, PyObject *args);

static PyObject *
XePPython_folder_get_message(PyObject *self, PyObject *args);

/* XetPan Python Module */

static PyMethodDef xetpan_module_def[] =
  {
    {
      "get_folder_list",
      XePPython_get_folder_list,
      METH_VARARGS,
      "Get the name list of all known folders"
    },
#if 0
    {
      "get_mainview_list",
      XePPython_get_mainview_list,
      METH_VARARGS,
      "Get the list of all open mainviews (pointers)"
    },
#endif
    {
      "update_folder",
      XePPython_update_folder,
      METH_VARARGS,
      "Update a folder. Need the reference of a folder as an argument."
    },
    {
      "condition_new",
      XePPython_condition_new,
      METH_VARARGS,
      "condition_new(ID, EXPRESSION): create a new condition"
    },
    {
      "condition_get",
      XePPython_condition_get,
      METH_VARARGS,
      "condition_get(ID): return the content of a condition."
    },
    {
      "action_new",
      XePPython_action_new,
      METH_VARARGS,
      "action_new(ID, COMMAND): create a new action"
    },
    {
      "action_get",
      XePPython_action_get,
      METH_VARARGS,
      "action_get(ID): return the content of an action."
    },
    {
      "rule_new",
      XePPython_rule_new,
      METH_VARARGS,
      "rule_new(ID, CONDITION_ID, ACTION_ID): create a new rule by combining existant condition and action."
    },
    {
      "rule_get",
      XePPython_rule_get,
      METH_VARARGS,
      "rule_get(ID): return the content of a rule."
    },
    {
      "folder_get_rules",
      XePPython_folder_get_rules,
      METH_VARARGS,
      "folder_get_rules(FOLDER_ID): return the list of processing rules of folder FOLDER_ID"
    },
    {
      "folder_add_rule",
      XePPython_folder_add_rule,
      METH_VARARGS,
      "folder_add_rule(FOLDER_ID, RULE_ID): add the processing rule RULE_ID to the list of processing rules of folder FOLDER_ID"
    },
    {
      "folder_remove_rule",
      XePPython_folder_remove_rule,
      METH_VARARGS,
      "folder_remove_rule(FOLDER_ID, RULE_ID): remove the processing rule RULE_ID from the list of processing rules of folder FOLDER_ID"
    },
    {
      "folder_get_uids",
      XePPython_folder_get_uids,
      METH_VARARGS,
      "folder_get_uids(FOLDER_ID): return a tuple containing the unique identifiers (UID) of the messages in the folder FOLDER_ID"
    },
    {
      "folder_get_message",
      XePPython_folder_get_message,
      METH_VARARGS,
      "folder_get_message(FOLDER_ID, MESSAGE_UID):"
    },
    {NULL, NULL, 0, NULL}
  };



int
a_XePPython_is_initialized(void)
{
  return Py_IsInitialized();
}

void
a_XePPython_initialize(void)
{
  PyObject *xetpan_module;

  Py_Initialize();

#ifdef USE_PYTHON_WITH_THREAD_SUPPORT
  PyEval_InitThreads();

  main_thread_state = PyThreadState_Get();
#endif

  /* Global XetPan variables */
  xetpan_module = Py_InitModule("xetpan", xetpan_module_def);

  PyRun_SimpleString("from xetpan import *");
  
#ifdef USE_PYTHON_WITH_THREAD_SUPPORT
  PyEval_ReleaseLock();
#endif
}


void
a_XePPython_finalize(void)
{
#ifdef USE_PYTHON_WITH_THREAD_SUPPORT
  a_XePPython_thread_main_acquire();
#endif
  Py_Finalize();
}

void
a_XePPython_execute(char *text)
{
#ifdef USE_PYTHON_WITH_THREAD_SUPPORT
  PyEval_AcquireLock();
  PyThreadState_Swap(main_thread_state);
#endif  
  
  PyRun_SimpleString(text);

#ifdef USE_PYTHON_WITH_THREAD_SUPPORT
  PyThreadState_Swap(NULL);
  PyEval_ReleaseLock();
#endif
}


#ifdef USE_PYTHON_WITH_THREAD_SUPPORT
void
a_XePPython_thread_execute(char *text, PyThreadState *thread_state)
{
  PyEval_AcquireLock();
  PyThreadState_Swap(thread_state);
  
  PyRun_SimpleString(text);

  PyThreadState_Swap(NULL);
  PyEval_ReleaseLock();
}

PyThreadState *
a_XePPython_thread_state_new(void)
{
  PyThreadState *thread_state;
  PyEval_AcquireLock();
  thread_state = PyThreadState_New(main_thread_state->interp);
  PyEval_ReleaseLock();
  return thread_state;
}

void
a_XePPython_thread_main_acquire(void)
{
  PyEval_AcquireLock();
  PyThreadState_Swap(main_thread_state);
}

void
a_XePPython_thread_main_release(void)
{
  PyThreadState_Swap(NULL);
  PyEval_ReleaseLock();
}
#endif

static PyObject *
XePPython_get_folder_list(PyObject *self, PyObject *args)
{
  PyObject *py_folder_ref;
  PyObject *py_folder_tuple;
  chash *folder_ref_hash_table = a_Folder_ref_hash_get();
  chashiter *iter;
  chashdatum key;
  int i;

  /* Need no arguments */
  if (!PyArg_ParseTuple(args, ""))
    return NULL;

  py_folder_tuple = PyTuple_New(chash_count(folder_ref_hash_table));

  for (i = 0, iter = chash_begin(folder_ref_hash_table); 
       iter;
       iter = chash_next(folder_ref_hash_table, iter), i++)
    {
      chash_key(iter, &key);
      py_folder_ref = PyString_FromString((char *)key.data);
      PyTuple_SetItem(py_folder_tuple, i, py_folder_ref);
    }
  return py_folder_tuple;  
}

#if 0
/* FIXME: mainviews is now a cinthash not a clist */
static PyObject *
XePPython_get_mainview_list(PyObject *self, PyObject *args)
{
  PyObject *py_uimain;
  PyObject *py_uimain_tuple;
  clist *mainviews = global_xetpan->ui->mainviews;
  clistiter *iter;
  int i;

  /* Need no arguments */
  if (!PyArg_ParseTuple(args, ""))
    return NULL;

  py_uimain_tuple = PyTuple_New(clist_count(mainviews));

  for (i = 0, iter = clist_begin(mainviews); 
       iter;
       iter = iter->next, i++)
    {
      py_uimain = PyCObject_FromVoidPtr(iter->data, NULL);
      PyTuple_SetItem(py_uimain_tuple, i, py_uimain);
    }
  return py_uimain_tuple;  
}
#endif

static PyObject *
XePPython_update_folder(PyObject *self, PyObject *args)
{
  const char *folder_ref;

  if (!PyArg_ParseTuple(args, "s", &folder_ref))
    return NULL;

  a_UI_update_folder(folder_ref);

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
XePPython_folder_get_rules(PyObject *self, PyObject *args)
{
  const char *folder_ref;
  XePFolder *folder;
  PyObject *py_rule_ref;
  PyObject *py_rule_tuple;
  clistiter *iter;
  XePProcessingRef *rule_ref;
  int i;

  /* Need no arguments */
  if (!PyArg_ParseTuple(args, "s", &folder_ref))
    return NULL;

  folder = a_Folder_get_from_ref(folder_ref);
  if (folder == NULL)
    {
      Py_INCREF(Py_None);
      return Py_None;
    }

  py_rule_tuple = PyTuple_New(clist_count(folder->rule_refs));

  for (i = 0, iter = clist_begin(folder->rule_refs); 
       iter;
       iter = iter->next, i++)
    {
      rule_ref = (XePProcessingRef *) iter->data;
      py_rule_ref = PyString_FromString(rule_ref->ref);
      PyTuple_SetItem(py_rule_tuple, i, py_rule_ref);
    }
  return py_rule_tuple;  
}

static PyObject *
XePPython_folder_add_rule(PyObject *self, PyObject *args)
{
  const char *folder_ref;
  const char *rule_ref;
  XePFolder *folder;

  if (!PyArg_ParseTuple(args, "ss", &folder_ref, &rule_ref))
    return NULL;

  folder = a_Folder_get_from_ref(folder_ref);
  if (folder)
    a_Folder_processing_rule_add(folder, rule_ref);

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
XePPython_folder_remove_rule(PyObject *self, PyObject *args)
{
  const char *folder_ref;
  const char *rule_ref;
  XePFolder *folder;

  if (!PyArg_ParseTuple(args, "ss", &folder_ref, &rule_ref))
    return NULL;

  folder = a_Folder_get_from_ref(folder_ref);
  if (folder)
    a_Folder_processing_rule_remove(folder, rule_ref);

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
XePPython_condition_new(PyObject *self, PyObject *args)
{
  const char *id;
  const char *expression;
  PyObject *py_result;
  int r;

  if (!PyArg_ParseTuple(args, "ss", &id, &expression))
    return NULL;

  r = a_Processing_condition_add_new(id, expression);

  py_result = PyInt_FromLong(r);

  return py_result;
}

static PyObject *
XePPython_condition_get(PyObject *self, PyObject *args)
{
  const char *id;
  XePProcessingCondition *condition;
  PyObject *py_result;

  if (!PyArg_ParseTuple(args, "s", &id))
    return NULL;

  condition = a_Processing_condition_get_from_ref(id);
  if (condition)
    {
      py_result = PyString_FromString(condition->expression);
      return py_result;
    }
  else
    {
      Py_INCREF(Py_None);
      return Py_None;
    }
}

static PyObject *
XePPython_action_new(PyObject *self, PyObject *args)
{
  const char *id;
  const char *command;
  PyObject *py_result;
  int r;

  if (!PyArg_ParseTuple(args, "ss", &id, &command))
    return NULL;

  r = a_Processing_action_add_new(id, command);

  py_result = PyInt_FromLong(r);

  return py_result;
}

static PyObject *
XePPython_action_get(PyObject *self, PyObject *args)
{
  const char *id;
  XePProcessingAction *action;
  PyObject *py_result;

  if (!PyArg_ParseTuple(args, "s", &id))
    return NULL;

  action = a_Processing_action_get_from_ref(id);
  if (action)
    {
      py_result = PyString_FromString(action->command);
      return py_result;
    }
  else
    {
      Py_INCREF(Py_None);
      return Py_None;
    }
}

static PyObject *
XePPython_rule_new(PyObject *self, PyObject *args)
{
  const char *id;
  const char *condition_ref;
  const char *action_ref;
  PyObject *py_result;
  int r;

  if (!PyArg_ParseTuple(args, "sss", &id, &condition_ref, &action_ref))
    return NULL;

  r = a_Processing_rule_add_new(id, condition_ref, action_ref);

  py_result = PyInt_FromLong(r);

  return py_result;
}

static PyObject *
XePPython_rule_get(PyObject *self, PyObject *args)
{
  const char *id;
  XePProcessingRule *rule;
  PyObject *py_result;

  if (!PyArg_ParseTuple(args, "s", &id))
    return NULL;

  rule = a_Processing_rule_get_from_ref(id);
  if (rule)
    {
      PyObject *py_rule_tuple;

      py_rule_tuple = PyTuple_New(2);
      py_result = PyString_FromString((rule->condition && rule->condition->id ?
				       rule->condition->id : ""));
      PyTuple_SetItem(py_rule_tuple, 0, py_result);
      py_result = PyString_FromString((rule->action && rule->action->id ?
				       rule->action->id : ""));
      PyTuple_SetItem(py_rule_tuple, 1, py_result);
      return py_rule_tuple;
    }
  else
    {
      Py_INCREF(Py_None);
      return Py_None;
    }
}

static PyObject *
XePPython_folder_get_uids(PyObject *self, PyObject *args)
{
  const char *folder_ref;
  XePFolder *folder;

  if (!PyArg_ParseTuple(args, "s", &folder_ref))
    return NULL;

  folder = a_Folder_get_from_ref(folder_ref);
  if (folder)
    {
      carray *msgs;
      int i;
      int count;
      PyObject *py_uid_tuple;
      PyObject *py_uid;
      mailmessage *msg;
      msgs = folder->ep_list->msg_tab;
      count = carray_count(msgs);
      py_uid_tuple = PyTuple_New(count);
      for (i = 0; i < count; i++)
	{
	  msg = carray_get(msgs, i);
	  if (msg->msg_uid)
	    {
	      py_uid = PyString_FromString(msg->msg_uid);
	      PyTuple_SetItem(py_uid_tuple, i, py_uid);
	    }
	}
      return py_uid_tuple;      
    }
  else
    {
      Py_INCREF(Py_None);
      return Py_None;
    }
}

static PyObject *
XePPython_folder_get_message(PyObject *self, PyObject *args)
{
  const char *uid;
  const char *folder_ref;
  XePFolder *folder;

  if (!PyArg_ParseTuple(args, "ss", &folder_ref, &uid))
    return NULL;
  
  folder = a_Folder_get_from_ref(folder_ref);
  if (folder)
    {
      XePMessage *message;
      char *data;
      PyObject *py_rfc2822;
      message = a_Message_new();
      message->uid = uid;
      message->folder = folder;
      a_Message_fetch_by_uid(message);
      message->uid = NULL;
      data = message->data;
      message->data = NULL;
      a_Message_unref(message);
      if (data)
	{
	  py_rfc2822 = PyString_FromString(data);
	  return py_rfc2822;
	}
    }
  Py_INCREF(Py_None);
  return Py_None;
}

#endif
